##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Asynchronous call engine.

$Id: client.py,v 1.8 2004/12/09 19:28:48 poster Exp $
"""

# To do:
# - on an exit (keyboard or system), try to stop polling but finish
#   processing remaining zope deferreds, if possible, at least just timing 
#   out
# - try to modify to use twistd so tools such as manhole work easily
# - catch POSKeyErrors too; what the heck.

# Debugging notes:
# strace -p pid on linux and ktrace -p pid (plus kdump) on os x/bsd should
# catch hanging processes.
# if I switch to using twistd then we can use manhole

import sys, getopt, logging, urlparse, StringIO, datetime
from ZODB.POSException import ConflictError
from ZEO.Exceptions import ClientDisconnected
from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse
from ZPublisher.BaseRequest import RequestContainer

from twisted.internet import reactor, defer
from twisted.python import failure

### the following values are set or modified by config:

max_conflict_resolution_attempts = 5
initial_retry_delay = 5
retry_exponential_backoff = 1.1
max_total_retry = 3600 # change to total max retry, 0 means no max
verbose_traceback = False

tool_path = None # this will hold the path to the Zope tool.  This variable
# is intentionally not used directly in this file, but that may change.

app = None # this will hold the Zope application getter

DB = None # this will hold the database

plugins = {} # the plugins available

### end configured

server_retries = [] # calls that need to be resumed once the ZEO server is 
                    # back

tool_retries = {} # calls that need to be resumed once a tool is back ({tool
                  # path: [calls]})

active = {} # {zope deferred id: (original timeout, zasync deferred)} of calls 
            # currently in process

server_retry_start = None
tool_retry_start = None

chores = [] # list of (function, args, kwargs) to be performed regularly by
# housekeeping

# helpers

def housekeeping(interval=4):
    global chores
    log = logging.getLogger('zasync')
    log.debug('housekeeping: performing %d chores', len(chores))
    for (f, args, kwargs) in chores:
        reactor.callLater(0, f, *args, **kwargs)
    reactor.callLater(interval, housekeeping, interval)

def getRequestApp(application):
    response = HTTPResponse(stdout=sys.stdout)
    request = HTTPRequest(
        sys.stdin,
        {'SERVER_NAME':'zasync_pseudo_server',
         'SERVER_PORT':'80',
         'REQUEST_METHOD':'GET'},
        response)
    return application.__of__(RequestContainer(REQUEST=request))

def logException(msg='', log=None):
    global verbose_traceback
    if log is None:
        log = logging.getLogger('zasync')
    get_transaction().abort()
    msg = 'Logged Exception.  Transaction aborted.  %s' % msg
    if verbose_traceback:
        res = failure.Failure()
        out = StringIO.StringIO()
        res.printDetailedTraceback(out)
        out = out.getvalue()
        log.error("%s\n\n%s\n\n%s", msg, out, res.getErrorMessage())
    else:
        log.error(msg, exc_info=True)
    return res

def scheduleServerRetry(call, *args, **kwargs):
    get_transaction().abort()
    global server_retries, initial_retry_delay, server_retry_start
    server_retries.append((call, args, kwargs))
    if len(server_retries) == 1:
        server_retry_start = datetime.datetime.now()
        logging.getLogger('zasync').critical(
            'disconnected from ZEO server; retrying.')
        # starting up a new backoff
        reactor.callLater(initial_retry_delay, 
                          retryServer, initial_retry_delay)
def is_connected():
    global DB
    storage = DB._storage
    call = getattr(storage, 'is_connected', None)
    if call is not None:
        return call()
    return True # we'll assume that non-ZEO storages (presumably only useful
    # for tests) are always connected

def retryServer(delay):
    global server_retries, retry_exponential_backoff
    global server_retry_start, max_total_retry
    if is_connected():
        server_retry_start = None
        delay = 0
        interval = 0.05
        while server_retries:
            call, args, kwargs = server_retries.pop(0)
            reactor.callLater(delay, call, *args, **kwargs)
            delay += interval
        logging.getLogger('zasync').info('reconnected to ZEO server')
    else:
        delay = pow(delay, retry_exponential_backoff)
        diff = datetime.datetime.now() - server_retry_start
        seconds = diff.seconds + 86400 * diff.days
        if (max_total_retry and seconds + delay > max_total_retry):
            logging.getLogger('zasync').critical(
                'disconnected from ZEO server for %d seconds, and maximum '
                'total retry reached.  Stopping.', seconds)
            raise SystemExit('ZEO server unreachable')
        else:
            logging.getLogger('zasync').critical(
                'disconnected from ZEO server for %d seconds; retrying.', 
                seconds)
            reactor.callLater(delay, retryServer, delay)

def scheduleToolRetry(path, call, *args, **kwargs):
    get_transaction().abort()
    global tool_retries, initial_retry_delay
    global tool_retry_start
    if path not in tool_retries:
        tool_retries[path] = [(call, args, kwargs)]
        logging.getLogger('zasync').critical(
            'can no longer find tool: /%s',
            '/'.join(path))
        tool_retry_start = datetime.datetime.now()
        reactor.callLater(initial_retry_delay,
                          retryTool, path, initial_retry_delay)
    else:
        tool_retries[path].append((call, args, kwargs))

def retryTool(path, delay):
    global retry_exponential_backoff, max_total_retry
    global tool_retry_start, app, DB
    application = None
    try:
        try:
            try:
                sync = DB._storage.sync # important
            except AttributeError:
                pass
            else:
                sync()
            application = app()
            tool = application.unrestrictedTraverse(path)
            get_transaction().commit()
        except ConflictError:
            logging.getLogger('zasync').info(
                'Received ConflictError while trying to retry tool; retrying.', 
                exc_info=True)
            get_transaction().abort()
            reactor.callLater(delay, retryTool, path, delay)
        except ClientDisconnected:
            logging.getLogger('zasync').critical(
                'ZEO server disconnected while trying to retry tool')
            get_transaction().abort()
            scheduleServerRetry(retryTool, path, delay)
        except (AttributeError, LookupError):
            get_transaction().abort()
            delay = pow(delay, retry_exponential_backoff)
            diff = datetime.datetime.now() - tool_retry_start
            seconds = diff.seconds + 86400 * diff.days
            if (max_total_retry and seconds + delay > max_total_retry):
                logging.getLogger('zasync').critical(
                    'cannot find tool /%s for %d seconds, and maximum retry '
                    'delay reached.  Stopping.', '/'.join(path), seconds)
                raise SystemExit('tool not found')
                delay = max_retry_delay
            else:
                logging.getLogger('zasync').critical(
                    'cannot find tool for %d seconds; retrying.',
                    seconds)
                reactor.callLater(delay, retryTool, path, delay)
        else:
            global tool_retries
            retries = tool_retries.pop(path)
            delay = 0
            interval = 0.05
            # schedule all the past calls
            found = False
            for call, args, kwargs in retries:
                reactor.callLater(delay, call, *args, **kwargs)
                delay += interval
                if not found and call is _setPlugins:
                    found = True
            if not found:
                # make sure tool (possibly new) knows about my plugins
                setPlugins(path) # returns a deferred, which we ignore
            logging.getLogger('zasync').info(
                'tool /%s found again; recommencing calls.', '/'.join(path))
    finally:
        if application is not None:
            application._p_jar.close()

def timeoutErrback(deferred):
    if not deferred.called:
        deferred.errback(defer.TimeoutError('Timed out.'))

def cancelDelayedCall(value, call):
    if call.active():
        call.cancel()
    return value

# pollZope schedules makeCall which schedules returnResult.  That's it, really.
def pollZope(path):
    "polls Zope to check for new calls, timeout changes, and expired calls"
    global active, app, DB, plugins
    application = None
    try:
        if is_connected():
            log = logging.getLogger('zasync')
            log.debug('polling %s', '/'.join(path))
            try:
                try:
                    sync = DB._storage.sync # important
                except AttributeError:
                    pass
                else:
                    sync()
                application = app()
                try:
                    tool = application.unrestrictedTraverse(path)
                except (AttributeError, LookupError):
                    scheduleToolRetry(path, pollZope, path)
                    return
                poll_interval = tool.poll_interval
                tool.heartbeat()
                delay = 0
                interval = 0.05
                for zopeDeferred in tool.acceptAll():
                    remainingSeconds = zopeDeferred.remainingSeconds()
                    if remainingSeconds <= 0:
                        # don't bother calling if it has already timed out
                        zopeDeferred.errback(
                            failure.Failure(defer.TimeoutError('Timed out.')))
                        continue
                    name, args, kwargs = zopeDeferred.getSignature()
                    call_info = plugins.get(name)
                    if call_info is None:
                        zopeDeferred.errback(failure.Failure(KeyError(name)))
                        continue
                    log.debug('got call for plugin %s (args %r; kwargs %r)', 
                              name, args, kwargs)
                    zopeDeferredId = zopeDeferred.key
                    timeout = zopeDeferred.timeout
                    calltimeout = call_info['timeout']
                    if timeout is None or calltimeout < timeout:
                        zopeDeferred.timeout = calltimeout
                    reactor.callLater(
                        delay, makeCall, 
                        path, zopeDeferredId, name, args, kwargs, 
                        remainingSeconds, timeout)
                    delay += interval
                bad = []
                for zopeDeferredId, (oldTimeout, deferred) in active.items():
                    zopeDeferred = tool.getDeferred(zopeDeferredId)
                    if zopeDeferred is None:
                        bad.append(zopeDeferredId)
                    elif (zopeDeferred.timeout < oldTimeout and 
                        deferred.timeoutCall.active()):
                        deferred.timeoutCall.reset(
                            max(zopeDeferred.remainingSeconds(), 0))
                for zopeDeferredId in bad:
                    del active[bad]
                get_transaction().commit()
            except ClientDisconnected:
                scheduleServerRetry(pollZope, path)
                return
            except ConflictError:
                get_transaction().abort()
                log.warning('ZODB ConflictError in pollZope.  Never give up.', 
                            exc_info=True)
                reactor.callLater(1, pollZope, path) # never give up
                return
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                logException(log=log)
            # XXX if we want the heartbeat interval to be more reliable then
            # we need more care
            reactor.callLater(poll_interval, pollZope, path)
        else:
            scheduleServerRetry(pollZope, path)
    finally:
        if application is not None:
            application._p_jar.close()

def returnResult(value, path, zopeDeferredId, error=False, count=0):
    global active, app
    if isinstance(value, failure.Failure):
        value.cleanFailure()
    log = logging.getLogger('zasync')
    log.debug('returnResult got value for %s (%s):\n\n%r', 
              zopeDeferredId, path, value)
    try:
        del active[zopeDeferredId]
    except KeyError:
        pass
    application = None
    try:
        if is_connected():
            try:
                application = app()
                try:
                    tool = getRequestApp(application).unrestrictedTraverse(path)
                except (AttributeError, LookupError):
                    scheduleToolRetry(
                        path, returnResult, value, path, zopeDeferredId, 
                        error, count)
                    return value
                zopeDeferred = tool.getDeferred(zopeDeferredId)
                if zopeDeferred is None:
                    return value
                if error:
                    call = zopeDeferred.errback
                else:
                    call = zopeDeferred.callback
                res = call(value)
                get_transaction().commit()
            except ConflictError:
                log.warning('ZODB ConflictError in returnResult', exc_info=True)
                get_transaction().abort()
                if count < max_conflict_resolution_attempts:
                    reactor.callLater(
                        count, returnResult, value, path, zopeDeferredId, 
                        error, count+1)
                else:
                    res = failure.Failure()
                    out = StringIO.StringIO()
                    res.printDetailedTraceback(out)
                    log.error(
                        'Too many ConflictErrors in returnResult: '
                        'giving up.\n\n%s', out.getvalue())
            except ClientDisconnected:
                scheduleServerRetry(
                    returnResult, value, path, zopeDeferredId, 
                    error, count)
                res = value
            except (KeyboardInterrupt, SystemExit):
                raise
            except: # give up.  Looks like a bug.  Log should help fix it.
                res = logException('Exception from Zope.', log=log)
            return res
        else:
            scheduleServerRetry(
                returnResult, value, path, zopeDeferredId, error, count)
    finally:
        if application is not None:
            application._p_jar.close()
    return res

def makeCall(path, zopeDeferredId, name, args, kwargs, remainingSeconds,
             timeout, returnResult=returnResult, count=0):
    # zopeDeferredId is an id of a deferred in the asynchronous tool
    global plugins
    log = logging.getLogger('zasync')
    log.debug('makeCall called for %s (%s)',
              zopeDeferredId, path)
    if is_connected():
        try:
            call_info = plugins[name]
            call = call_info['callable']
        except KeyError:
            res = failure.Failure()
            log.error('Called plugin %s does not exist', name)
        else:
            log.debug('called %s', name)
            try:
                if call_info['zope_aware']:
                    res = call((path, zopeDeferredId), *args, **kwargs)
                else:
                    res = call(*args, **kwargs)
                get_transaction().commit() # just in case plugin touched Zope:
                # please do not!  Use zope_exec, or follow that pattern...
                # If you do, and you start a conflict error, it may throw
                # the timeout check off.  Just don't, and regard this as 
                # paranoia.
            except ConflictError:
                log.warning('ZODB ConflictError in makeCall', exc_info=True)
                get_transaction().abort()
                if count < max_conflict_resolution_attempts:
                    reactor.callLater(
                        count, makeCall, path, zopeDeferredId, 
                        name, args, kwargs, remainingSeconds, timeout, 
                        returnResult, count+1)
                    return
                else:
                    res = failure.Failure()
                    out = StringIO.StringIO()
                    res.printDetailedTraceback(out)
                    log.error(
                        'Too many ConflictErrors in makeCall: '
                        'giving up.\n\n%s', out.getvalue())
                    # now we use that res (Failure)
            except ClientDisconnected:
                scheduleServerRetry(
                    makeCall, path, zopeDeferredId, name, args, kwargs, 
                    remainingSeconds, timeout, returnResult, count)
                return
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                get_transaction().abort()
                res = failure.Failure()
        if isinstance(res, defer.Deferred):
            res.timeoutCall = reactor.callLater(
                max(remainingSeconds, 0), 
                timeoutErrback, res)
            res.addBoth(cancelDelayedCall, res.timeoutCall)
            active[zopeDeferredId] = (timeout, res)
            res.addCallbacks(
                returnResult, returnResult,
                callbackArgs=(path, zopeDeferredId),
                errbackArgs=(path, zopeDeferredId, True))
        elif isinstance(res, failure.Failure):
            res.cleanFailure()
            reactor.callLater(0, returnResult, res, path, zopeDeferredId, True)
        else:
            reactor.callLater(0, returnResult, res, path, zopeDeferredId)
    else:
        scheduleServerRetry(
            makeCall, path, zopeDeferredId, name, args, kwargs, 
            remainingSeconds, timeout, returnResult, count)

def stop(ignored=None):
    raise SystemExit()

def setPlugins(path):
    d = defer.Deferred()
    reactor.callLater(0, _setPlugins, d, path)
    return d

def _setPlugins(deferred, path, count=0):
    global app, plugins, max_conflict_resolution_attempts
    log = logging.getLogger('zasync')
    application = None
    try:
        try:
            application = app()
            try:
                tool = application.unrestrictedTraverse(tool_path)
            except (AttributeError, LookupError):
                scheduleToolRetry(path, _setPlugins, deferred, path, count)
                return
            else:
                tool.setPlugins(
                    [(n, p['description']) for n, p in plugins.items()])
                get_transaction().commit()
        except ConflictError:
            get_transaction().abort()
            if count == max_conflict_resolution_attempts-1:
                log.critical(
                    "Too many conflicts trying to setPlugins!", exc_info=True)
                deferred.errback(failure.Failure())
            else:
                count += 1
                log.info("Conflict error %d trying to setPlugins", count)
                reactor.callLater(count, _setPlugins, deferred, path, count)
        except ClientDisconnected:
            scheduleServerRetry(_setPlugins, deferred, path, count)
            return
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            get_transaction().abort()
            log.critical("Unexpected error: probably serious", exc_info=True)
            deferred.errback(failure.Failure())
        else:
            deferred.callback(path)
    finally:
        if application is not None:
            application._p_jar.close()

def handlePastCalls(path, count=0):
    global app
    log = logging.getLogger('zasync')
    application = None
    try:
        try:
            application = app()
            try:
                tool = application.unrestrictedTraverse(path)
            except (AttributeError, LookupError):
                scheduleToolRetry(path, handlePastCalls, path, count)
                return
            else:
                info = [(d.getSignature(), d.key, 
                         d.remainingSeconds(), d.timeout) for d in 
                        tool.getAcceptedCalls()]
                get_transaction().commit()
        except ConflictError:
            get_transaction().abort()
            if count==max_conflict_resolution_attempts-1:
                log.critical(
                    "Too many conflicts trying to handlePastCalls!", 
                    exc_info=True)
                raise SystemExit("Too many conflicts trying to handlePastCalls")
            else:
                count+=1
                log.info("Conflict error %d trying to handlePastCalls", count)
                reactor.callLater(count, handlePastCalls, path, count)
        except ClientDisconnected:
            scheduleServerRetry(handlePastCalls, path, count)
            return
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            get_transaction().abort()
            log.critical("Unexpected error: probably serious", exc_info=True)
            deferred.errback(failure.Failure())
        else:
            for (name, args, kwargs), key, remainingSeconds, timeout in info:
                try:
                    call_info = plugins[name]
                except KeyError:
                    retry = False
                else:
                    retry = call_info.get('retry', False)
                if retry and remainingSeconds > 0:
                    reactor.callLater(
                        0, makeCall, tool_path, key, name, args, kwargs,
                        remainingSeconds, timeout)
                else:
                    reactor.callLater(
                        0,
                        returnResult, 
                        failure.Failure(
                            defer.TimeoutError(
                                'zasync was disconnected (now reconnected)')),
                        tool_path,
                        key,
                        error=True)
            log.debug('scheduling first pollZope')
            reactor.callLater(0, pollZope, path)
            log.debug('scheduling first housekeeping')
            reactor.callLater(0, housekeeping)
    finally:
        if application is not None:
            application._p_jar.close()

def run(path=None):
    # to be called after config.initialize
    global tool_path, app
    if path is None:
        path = tool_path
    tool_path = path = tuple(path)
    d = setPlugins(path)
    d.addCallbacks(handlePastCalls, stop)
    d.addErrback(stop)
    log = logging.getLogger('zasync')
    log.debug('beginning reactor')
    try:
        reactor.run()
    finally:
        application = None
        try:
            get_transaction().abort()
            application = app()
            try:
                tool = application.unrestrictedTraverse(tool_path)
                tool.setPlugins(())
                get_transaction().commit()
            except (AttributeError, KeyError, ConflictError, ClientDisconnected):
                get_transaction().abort()
        finally:
            if application is not None:
                application._p_jar.close()
            logging.getLogger('zasync').critical('Shutting down')
            logging.shutdown()
