##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""zconfig helpers

$Id: config.py,v 1.3 2004/10/28 01:31:27 poster Exp $
"""

import os, sys, types, traceback, re, sets
import ZConfig

def root_handler(config):
    """Set up the environment and generally do almost everything;
    this was combined from datatypes and configs in Zope gradually as
    necessary elements emerged; it should probably be separated again
    at some point in the future."""
    
    # This is in two parts: set up Zope, then set up zasync.  Unsurprisingly,
    # it takes a lot more to set up Zope.
    
    # ...but first we'll set up logging
    import zLOG 
    zLOG._call_initialize = 0
    config.eventlog()
    config.zasync()
    config.zasync_plugins.name="zasync.plugins"
    config.zasync_plugins()
    
    ###############
    # Set up Zope #
    ###############

    # Set environment variables
    for k,v in config.environment.items():
        os.environ[k] = v
    
    # set up Zope as minimally as possible while still getting it to work
    
    # Add directories to the pythonpath; always insert instancehome/lib/python
    instancelib = os.path.join(config.instancehome, 'lib', 'python')
    if instancelib not in config.path:
        config.path.append(instancelib)
    path = config.path[:]
    path.reverse()
    for dir in path:
        sys.path.insert(0, dir)

    # Add any product directories not already in Products.__path__.
    # Directories are added in the order they are mentioned
    # Always insert instancehome.Products

    instanceprod = os.path.join(config.instancehome, 'Products')
    if instanceprod not in config.products:
        config.products.append(instanceprod)
    
    import Products
    L = []
    for d in config.products + Products.__path__:
        if d not in L:
            L.append(d)
    Products.__path__[:] = L
    
    import App, App.config
    par = os.path.dirname
    config.softwarehome = par(par(os.path.abspath(App.__file__)))
    config.zopehome = par(par(d))
    config.debug_mode = False # configurable?
    config.enable_product_installation = False
    if getattr(config, 'clienthome', None) is None:
        config.clienthome = os.path.join(config.instancehome, 'var')
    App.config.setConfiguration(config)
    
    import AccessControl
    AccessControl.setImplementation(
        config.security_policy_implementation)
    AccessControl.setDefaultBehaviors(
        not config.skip_ownership_checking,
        not config.skip_authentication_checking)
    
    import OFS.Application
    OFS.Application.import_products()
    
    from ZODB import ZApplication
    from AccessControl.SecurityManagement import getSecurityManager
    from AccessControl.SecurityManagement import newSecurityManager
    from AccessControl.SecurityManagement import noSecurityManager
    import AccessControl.User
    import Globals
    from Zope import ClassFactory

    Globals.DatabaseVersion='3'
    
    DB = config.dbtab.getDatabase('/', is_root=1)
    DB.setClassFactory(ClassFactory.ClassFactory)

    # "Log on" as system user
    newSecurityManager(None, AccessControl.User.system)

    # Set up the "app" object that automagically opens
    # connections
    app = ZApplication.ZApplicationWrapper(
        DB, 'Application', OFS.Application.Application, (),
        Globals.VersionNameName)

    # Initialize the app object
    application = app()
    OFS.Application.initialize(application)
    # ...application is closed below

    # "Log off" as system user
    noSecurityManager()
    
    #################
    # Set up zasync #
    #################
    
    from zasync import client
    
    client.DB = DB
    client.app = app
    
    path = client.tool_path = tuple(filter(None, config.target.split('/')))
    config.target_path = path
    try:
        tool = application.unrestrictedTraverse(path)
    except (AttributeError, LookupError):
        raise ZConfig.ConfigurationError(
            "target could not be found: %s" % path)
    
    application._p_jar.close()
    
    # configure retry behavior
    client.max_conflict_resolution_attempts = (
        max(config.max_conflict_resolution_attempts, 0))
    client.initial_retry_delay = (
        max(config.initial_retry_delay, 1))
    client.retry_exponential_backoff = (
        max(config.retry_exponential_backoff, 1))
    client.max_total_retry = max(config.max_total_retry, 0)
    
    # configure traceback verbosity
    client.verbose_traceback = config.verbose_traceback
    
    # load in plugins.  We don't want to calculate the callables until 
    # this root handler because the python environment is not set up until here.
    for plugin in config.plugins:
        # based off of 
        # http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/223972
        fullFuncName = plugin['handler']
        lastDot = fullFuncName.rfind(u".")
        if lastDot==-1:
            raise ZConfig.ConfigurationError(
                'plugin not found (code expects to find a dot in path): %s' %
                fullFuncName)
        funcName = fullFuncName[lastDot + 1:]
        modulePath = fullFuncName[:lastDot]
        try:
            aMod = sys.modules[modulePath]
            if not isinstance(aMod, types.ModuleType):
                raise KeyError
        except KeyError:
            try:
                aMod = __import__(modulePath, globals(), locals(), [''])
            except ImportError:
                tb = ''.join(traceback.format_exception(*sys.exc_info()))
                raise ZConfig.ConfigurationError(
                    'error importing %s: \n%s' % (fullFuncName, tb))
            sys.modules[modulePath] = aMod
        func = getattr(aMod, funcName)
        if not callable(func):
            raise ZConfig.ConfigurationError(
                'identified plugin must be callable: %s' % fullFuncName)
        plugin['callable'] = func
        plugin['description'] = (
            plugin.get('description') or func.__doc__ or func.__name__)
        client.plugins[plugin['name']] = plugin

# simple datatypes
def plugin(section):
    return {
        'name':section.getSectionName(),
        'handler':section.handler,
        'description':getattr(section, 'description', None),
        'retry':section.retry,
        'timeout':max(getattr(section, 'timeout', sys.maxint), 10),
        'zope_aware':section.zope_aware,
        }

def root_section(section):
    if not section.databases:
        section.databases = []
    mount_factories = {} # { name -> factory}
    mount_points = {} # { virtual path -> name }
    dup_err = ('Invalid configuration: ZODB databases named "%s" and "%s" are '
               'both configured to use the same mount point, named "%s"')
    for database in section.databases:
        points = database.getVirtualMountPaths()
        name = database.config.getSectionName()
        mount_factories[name] = database
        for point in points:
            if mount_points.has_key(point):
                raise ConfigurationError(dup_err % (mount_points[point],
                                                    name, point))
            mount_points[point] = name
    from DBTab.DBTab import DBTab
    section.dbtab = DBTab(mount_factories, mount_points)
    
    s = sets.Set()
    for plugin in section.plugins:
        name = plugin['name']
        if name in s:
            raise ConfigurationError(
                'plugin name duplicated: %s' % name)
            s.add(name)
    
    return section

_ident_re = "[_a-zA-Z][_a-zA-Z0-9]*"
class DottedNameFunctionConversion(
    ZConfig.datatypes.RegularExpressionConversion):
    
    def __init__(self):
        ZConfig.datatypes.RegularExpressionConversion.__init__(
            self, r"%s(?:\.%s)+" % (_ident_re, _ident_re))
DottedNameFunctionConversion = DottedNameFunctionConversion()

class LoggerFactory:
    """
    A factory used to create loggers while delaying actual logger
    instance construction.  We need to do this because we may want to
    reference a logger before actually instantiating it (for example,
    to allow the app time to set an effective user).  An instance of
    this wrapper is a callable which, when called, returns a logger
    object.
    """
    def __init__(self, section):
        self.name = section.getSectionName()
        self.level = section.level
        self.handler_factories = section.handlers
        self.propagate = section.propagate
        self.resolved = None
        self.section = section

    def __call__(self):
        if self.resolved is None:
            # set the logger up
            import logging
            logger = logging.getLogger(self.name)
            logger.handlers = []
            logger.propagate = self.propagate
            logger.setLevel(self.level)
            for handler_factory in self.handler_factories:
                handler = handler_factory()
                logger.addHandler(handler)
            self.resolved = logger
        return self.resolved

### the circus master that runs the configuration

def initialize(conffile=None):
    mydir = os.path.dirname(__file__)
    schema = ZConfig.loadSchema(os.path.join(mydir, 'schema.xml'))
    if conffile is None: conffile = sys.argv[1]
    conf, handler = ZConfig.loadConfig(schema, conffile)
    handler({'root_handler':root_handler})
    return conf
