##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""interfaces (call manager and zope deferred)

$Id: interfaces.py,v 1.1.1.1 2004/10/10 23:37:06 poster Exp $
"""

from Interface import Interface, Attribute

class IZopeDeferred(Interface):

    result = Attribute(
        """The result returned from the asynchronous worker.  Initializes to 
        None.  Because None is a possible vaild result from a plugin, this does
        not indicate the state of the deferred: use getState or getRawState.
        
        If you would like to also be protected from timeouts caused by the
        asynchronous worker being off-line, use getResult.""")

    failure = Attribute(
        """The failure returned from the asynchronous worker.  Initializes to 
        None.  
        
        If you would like to also be protected from timeouts caused by the
        asynchronous worker being off-line, use getResult.""")
    
    original_result = Attribute(
        "The result, not affected by callbacks")
    
    original_failure = Attribute(
        "The original failure, not affected by callbacks")

    creation_date = Attribute(
        """A standard Python datetime.datetime indicating the creation date
        of the deferred""")
    
    key = Attribute(
        """The value that can be used to obtain this deferred from the 
        asynchronous call manager via getDeferred""")
    
    local_key = Attribute(
        """The value that can be used to obtain this deferred from the
        asynchronous call manager via getSessionDeferred""")
    
    def getSignature():
        """Return the signature of the zasync call: a tuple of the zasync 
        plugin name, the ordered arguments, and the keyword arguments."""
    
    def getState():
        """returns 'success', 'failure', or None, indicating whether the 
        deferred has received a success result, a failure result, or no result
        yet.  If the deferred has not received any response for an 
        implementation-specific interval after the deferred's timeout, returns
        'failure'.  If the deferred eventually does receive a response, the
        state may change to reflect the true response."""
    
    def getRawState():
        """Same as getState, except that no timeout check occurs, so the value
        is always exactly what has been received from the asynchronous worker.
        """
    
    def getValue():
        """If deferred has been called with a success, returns result; if 
        deferred has been called with a failure, returns failure; if deferred
        has not yet been called and an implementation-specific interval after
        the deferred's timeout has passed, returns a failure; otherwise raises
        a RuntimeError."""
    
    def getRawValue():
        """If deferred has been called with a success, returns result; if 
        deferred has been called with a failure, returns failure; otherwise 
        raises a RuntimeError."""
    
    def getErrorMessage():
        """If getState returns 'failure', tries to convert the failure to an
        error message using the failure getErrorMessage API.  If not a failure,
        raises RuntimeError."""
    
    def callback(result):
        """stores result, changes state, and fires the first callback of the
        callback/errback chain.  If it has been called before, raises 
        RuntimeError."""
    
    def errback(failure):
        """stores failure, changes state, and fires the first errback of the
        callback/errback chain.  If it has been called before, raises 
        RuntimeError."""
    
    def addCallback(callback):
        """Convenience method for addCallbacks(callback)."""
    
    def addErrback(errback):
        """Convenience method for addCallbacks(errback=errback)."""

    def addCallbacks(callback=None, errback=None):
        """add a callback, an errback, or both.  Passing neither raises a
        ValueError.  These calls may either be a basestring that will be
        evaluated as a TALES Expression; or a pair (list or tuple) of 
        (basestring, basestring) that indicate, in order, the name that the
        result should be stored as and the expression that should be 
        evaluated as a TALES expression (a parallel to the behavior of 
        tal:define).  Expressions have the following locally defined names 
        available to them:
        
        * 'nothing' is None;
         
        * 'user' is the current user;
         
        * 'userhome' is the container of the user's acl_users;
         
        * 'tool' is the asynchronous call manager;
         
        * 'root' is the Zope "physical" root;
         
        * 'deferred' is the current deferred object;
         
        * 'modules' is the standard Zope secure module loader;
         
        * 'request' is the request object at the time of the callback, which
          is virtually always an "artificial" request, not associated with
          a true browser request;
         
        * 'result' begins as the result of a successful callback, or as None
          if the deferred results in a failure, and then becomes any 
          non-failure result of previous callbacks or errbacks--see 
          deferred.original_result if you want the unambiguous initial value;
         
        * 'failure' begins as the failure of an unsuccessful callback, or as
          None if the deferred succeeds, and then becomes any 
          failure result of previous callbacks or errbacks--see 
          deferred.original_failure if you want the unambiguous initial value;
         
        * 'results' is a dictionary of any value stored using the 
          tal:define-inspired syntax of the pair described above."""
    
    def setTimeout(seconds):
        """set a timeout on a deferred.  This should be regarded as a rough 
        guide.  seconds must be >= 0, and if a timeout has been set for this
        deferred previously, seconds can only be <= the previous timeout.
        The deferred currently includes no way to cancel a timeout.  The 
        timeout value is calculated from the deferred's creation_date.
        seconds will be converted to an integer using standard casting (i.e.,
        "seconds = int(seconds)".
        
        The Twisted deferred API that inspired this variation has deprecated
        the equivalent call (also called "setTimeout").  The Zope deferred
        probably will continue to support this call because of its convenience.
        """
    def remainingSeconds():
        """returns number of seconds until timeout (positive integer), or 
        number of seconds since timeout (negative integer)"""
    
    # below are excerpts from the standard Zope Owned API. Zope deferreds use 
    # the deferred owner to specify the security privileges under which
    # callbacks and errbacks are run.
                                                                            
    def changeOwnership(user, recursive=0):
        """Change the ownership to the given user.  If 'recursive' is
        true then also take ownership of all sub-objects, otherwise
        sub-objects retain their ownership information.  'recursive' is not
        important for Zope deferreds.
        
        A representative method of the standard Zope "Owned" API: see 
        lib/python/AccessControl/Owned.py in the standard Zope distribution
        for the code.  Zope deferreds use the deferred owner to specify the
        security privileges under which callbacks and errbacks are run."""
    
    def getWrappedOwner():
        """Get the owner, as described in getOwnerTuple, modestly wrapped in 
        the user folder.  If the object is not owned, return None.  If the 
        owner's user database doesn't exist, return Nobody.  If the owner ID 
        does not exist in the user database, return Nobody.
        
        A representative method of the standard Zope "Owned" API: see 
        lib/python/AccessControl/Owned.py in the standard Zope distribution
        for the code.  Zope deferreds use the deferred owner to specify the
        security privileges under which callbacks and errbacks are run."""
    
    def getOwnerTuple():
        """Return a tuple, (userdb_path, user_id) for the owner.  Ownership 
        can be acquired, but only from the containment path.  If unowned, 
        return None.
        
        A representative method of the standard Zope "Owned" API: see 
        lib/python/AccessControl/Owned.py in the standard Zope distribution
        for the code.  Zope deferreds use the deferred owner to specify the
        security privileges under which callbacks and errbacks are run."""

class IAbstractAsynchronousCallManager(Interface):
    """The current zeo-client-backed approach isn't the only one: we might be
    able to do this same API in the same Zope process, if that becomes
    desirable, for instance.  This abstract interface indicates the calls that
    should be maintained."""
    
    def listPlugins():
        """show plugin names available.  key is name, value is description.
        If no plugins, no engine should be expected (i.e., the tool will
        not work).  This part of the API is particularly unstable."""
    
    def putCall(_plugin, *args, **kwargs):
        """Make a call to the plugin named by _plugin, with associated args
        and kwargs.  Currently the only sanity check is to confirm that the 
        plugin name is available: no checks on the signature are performed.
        Importantly, arguments *must not* be persistent objects.  This check
        is also currently not enforced, but is very important.  Returns a
        zope deferred for this call, from which you can get the key and with
        which you can set callbacks, errbacks, and timeouts.
        """
    
    def putSessionCall(_plugin, *args, **kwargs):
        """As with putCall, except the deferred key includes the current user's 
        browser identifier, as defined by the standard Zope session machinery.
        """
    
    def getDeferred(d_id, default=None):
        """Given a deferred key, try to return the associated deferred, or else
        return the default."""
    
    def getSessionDeferred(d_id, default=None):
        """Given a deferred key or local_key, try to return the associated 
        deferred, or else return the default.  If a key is used, the key
        must include the browser id of the current user's session, or else
        raises ValueError."""
    
    def __len__():
        """returns the total number of deferreds currently stored by the tool.
        """
    
    def __nonzero__():
        "returns True, no matter what the len is"
    
    # private to deferreds
    
    def resolve(deferred):
        """Informs the asynchronous tool that the deferred has been resolved,
        and is no longer new or accepted (==pending)."""

class IZasyncAsynchronousCallManager(IAbstractAsynchronousCallManager):
    
    # public
    
    rotation_period = Attribute(
        """The rotation period in seconds for the resolved deferred cache
        """)
        
    poll_interval = Attribute(
        """The approximate period of polling for the zasync process.
        """)

    def getNextCacheRotation():
        """Return the next datetime.datetime when the resolved deferred cache 
        will rotate."""
    
    def resetNextCacheRotation(clear=False):
        """Reset the next cache rotation to (now + rotation_period), or if 
        clear is True, to never (until another call is made and the cache
        rotation time is set again)."""
    
    # private to zasync
    
    def setPlugins(plugins):
        """The Zasync engine should pass a sequence of available plugins,
        in which each plugin is represented by a pair of (name, description),
        when it starts; and attempt to set an empty tuple when in stops.
        This part of the API is particularly unstable."""
    
    def acceptAll():
        """Get all newly-added calls and move them to accepted status."""
    
    def getAcceptedCalls():
        """list the accepted calls."""
    
    def heartbeat():
        """Called every poll_interval (approximately).  Do work as needed;
        currently handles cache rotation."""
