##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Asynchronous call manager

A place to hang asynchronous calls and get their results

$Id: manager.py,v 1.7 2004/12/06 20:17:15 poster Exp $
"""

from types import NoneType
import datetime, random, sys, os, logging, sets

from twisted.internet import defer
from twisted.python import failure

from Persistence import Persistent
from ZODB.POSException import ConflictError
from ZEO.Exceptions import ClientDisconnected
from Globals import InitializeClass, Persistent
from Acquisition import aq_base, aq_inner, aq_parent
from AccessControl.Owned import ownerInfo
from AccessControl import ClassSecurityInfo, Unauthorized, \
    getSecurityManager
from AccessControl.SecurityManagement import newSecurityManager, \
    setSecurityManager
from Persistence import PersistentMapping
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from BTrees import OOBTree
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PageTemplates.Expressions import getEngine, SecureModuleImporter
from Products.Sessions.BrowserIdManager import BROWSERID_MANAGER_NAME
from AccessControl.SecurityInfo import allow_class

import permissions, bforests, interfaces

UNCALLED = 0
CALLED = 1
FAILURE = -1
state_name_map = {UNCALLED: None, CALLED: 'success', FAILURE: 'failure'}

# try to make failure.Failure available to restricted python
allow_class(failure.Failure) # in particular, this should allow access to
# 'getErrorMessage', 'getBriefTraceback', 'getTraceback', and 'trap'.

def iterable(value):
    try:
        iter(value)
    except TypeError:
        return False
    else:
        return True

def countable(value):
    try:
        len(value)
    except TypeError:
        return False
    else:
        return True

class Reference:
    
    def __init__(self, obj):
        try:
            self._path = obj.getPhysicalPath()
        except (AttributeError, KeyError):
            self._path = None
        self._repr = repr(obj)
    
    def dereference(self, context):
        return context.restrictedTraverse(self._path, None)
    
    def __repr__(self):
        return self._repr
allow_class(Reference)

class StandIn:
    
    data = klass = contents = None
    
    def __init__(self, obj, depth, contents=None):
        d = getattr(obj, '__dict__', None)
        if d is not None:
            self.data = sanitize(d, depth)
        self.klass = obj.__class__
        self.contents = contents
        self._repr = repr(obj)
    
    def __getattr__(self, name):    
        if self.data is not None and self.data is not MAX_DEPTH_MARKER:
            try:
                return self.data[name]
            except KeyError:
                pass
        raise AttributeError(name)
    
    def __repr__(self):
        return self._repr
allow_class(StandIn)

MAX_DEPTH = 4

class MAX_DEPTH_MARKER: pass # singleton

def sanitize(value, depth=0, max_depth=MAX_DEPTH):
    if depth >= max_depth:
        return MAX_DEPTH_MARKER
    depth += 1
    if aq_base(value) is not value or isinstance(value, Persistent):
        if (aq_parent(value) is not None and 
            getattr(aq_base(value), 'getPhysicalPath', None) is not None):
            value = Reference(value)
        else:
            value = StandIn(value, depth) # XXX could theoretically do a better 
            # job for BTrees, but don't think I really want to make sending
            # those across the wire all that easy
    elif getattr(value, '__dict__', None) is not None:
        value = StandIn(value, depth) # XXX not ideal for subclasses of built 
        # in types, but oh well
    elif isinstance(value, (list, tuple, sets.Set)):
        val_type = type(value)
        contents = [sanitize(item, depth) for item in value]
        if val_type is list:
            value = contents
        else:
            try:
                value = val_type(contents)
            except TypeError:
                value = StandIn(value, depth, contents)
    elif isinstance(value, dict):
        contents = [
            (sanitize(k, depth), sanitize(v, depth)) for k, v in value.items()]
        val_type = type(value)
        try:
            value = val_type(contents)
        except TypeError:
            value = StandIn(value, depth, contents)
    elif not isinstance(value, basestring) and iterable(value):
        contents = None
        if countable(value):
            contents = [sanitize(item, depth) for item in value]
        value = StandIn(value, depth, contents)
    return value

# Expression class copied from CMFCore/Expression.py
class Expression (Persistent):
    text = ''
    _v_compiled = None

    security = ClassSecurityInfo()

    def __init__(self, text):
        self.text = text
        self._v_compiled = getEngine().compile(text)

    def __call__(self, econtext):
        compiled = self._v_compiled
        if compiled is None:
            compiled = self._v_compiled = getEngine().compile(self.text)
        res = compiled(econtext)
        if isinstance(res, Exception):
            raise res
        return res

class Deferred(SimpleItem):
    
    result = failure = original_result = original_failure = None
    key = local_key = resolution_date = None
    timeout = 60 * 60 * 24 # one day; zasync plugins generally constrain 
    # timeouts more strictly
    

    # keep webdav interface off deferreds and tool
    __implements__ = (interfaces.IZopeDeferred,)
    
    security = ClassSecurityInfo()
    
    def __init__(self, plugin, args, kwargs):
        self.creation_date = datetime.datetime.now()
        self.__signature = (plugin, args, kwargs)
        self.__callbacks = ()
        self.raw_state = UNCALLED # raw_state is not reliable.
        # Use getState.
    
    security.declareProtected(permissions.View, 'getSignature')
    def getSignature(self):
        return self.__signature
    
    security.declareProtected(permissions.View, 'getState')
    def getState(self):
        "returns 'success', 'failure', or None"
        state = self.raw_state
        if state==UNCALLED and (
            self.remainingSeconds() < -(self.aq_parent.poll_interval*3)):
            state=FAILURE
        global state_name_map
        return state_name_map[state]
    
    security.declareProtected(permissions.View, 'getRawState')
    def getRawState(self):
        "returns 'success', 'failure', or None"
        global state_name_map
        return state_name_map[self.raw_state]
    
    security.declareProtected(permissions.View, 'getValue')
    def getValue(self):
        state = self.raw_state
        if state==UNCALLED:
            if (self.remainingSeconds() < -(self.aq_parent.poll_interval*3)):
                return failure.Failure(defer.TimeoutError('Timed out.'))
            raise RuntimeError("Deferred still pending")
        if state==CALLED:
            return self.result
        else:
            return self.failure
    
    security.declareProtected(permissions.View, 'getRawValue')
    def getRawValue(self):
        state = self.raw_state
        if state==UNCALLED:
            raise RuntimeError("Deferred still pending")
        if state==CALLED:
            return self.result
        else:
            return self.failure
    
    security.declareProtected(permissions.View, 'getErrorMessage')
    def getErrorMessage(self):
        state = self.getState()
        if state != 'failure':
            raise RuntimeError("No failure state.", state or 'pending')
        value = self.getValue()
        return value and value.getErrorMessage() or 'Unknown error.'
    
    def __repr__(self):
        classname = type(self).__name__
        if aq_parent(self) is not None:
            location = " in %s" % '/'.join(self.getPhysicalPath())
        else:
            location = ""
        return "<%s (key %r)%s at %s>" % (
            classname, self.key, location, id(aq_base(self)))
    
    def __call_all(self):
        callbacks = self.__callbacks
        assert self.raw_state != UNCALLED
        res = (self.raw_state == FAILURE and self.failure or self.result)
        self.resolution_date = datetime.datetime.now()
        if not callbacks:
            return res
        user = self.getWrappedOwner()
        tool = aq_parent(self)
        root = self.getPhysicalRoot()
        results = {}
        old_sm = getSecurityManager()
        newSecurityManager(None, user)
        context_dict = {
            'nothing': None, # provided automatically but here for clarity
            'user': user,
            'userhome': aq_parent(aq_inner(aq_parent(aq_inner(user)))),
            'tool': tool,
            'root': root,
            'deferred': self,
            'here': self,
            'modules': SecureModuleImporter,
            'request': root.REQUEST,
            'result': self.result,
            'failure': self.failure,
            'original': res,
            'results': results,
            }
        try:
            callbacks = list(callbacks)
            self.__callbacks = ()
            while callbacks:
                callback, errback = callbacks.pop(0)
                if self.raw_state == CALLED:
                    call = callback
                else:
                    assert self.raw_state == FAILURE
                    call = errback
                name = None
                if call is None:
                    continue # no-op
                elif isinstance(call, tuple):
                    name, call = call
                context = getEngine().getContext(context_dict) # XXX looks like
                # we need to create a new one of these each time in order to 
                # make the change to result or failure stick.  Before tried
                # context.contexts['result'] = res (and similar for failure)
                # to change, but no success.
                try:
                    res = call(context)
                except ((ConflictError, KeyboardInterrupt, 
                         SystemExit, ClientDisconnected)):
                    raise
                except:
                    res = failure.Failure()
                else:
                    if name is not None:
                        results[name] = res
                if isinstance(res, failure.Failure):
                    res.cleanFailure() # remove object references
                    self.failure = context_dict["failure"] = res
                    self.raw_state = FAILURE
                else:
                    self.result = context_dict["result"] = res
                    self.raw_state = CALLED
        finally:
            setSecurityManager(old_sm)
        return res
        # manage transactions and retries externally (zasync)
    
    security.declarePrivate('callback')
    def callback(self, result):
        self.aq_inner.aq_parent.resolve(self)
        if self.raw_state != UNCALLED:
            # we don't want to abort the transaction because there's a good 
            # chance that we *need* to have the manager resolve this deferred.
            # Therefore, we return a failure rather than raise a transaction.
            # We do not call the callbacks.
            logging.getLogger('zasync').error(
                "Zope deferred %r has already been called" % (self.key,))
            return failure.Failure(
                RuntimeError("Deferred has already been called"))
        self.raw_state = CALLED
        self.original_result = self.result = result
        return self.__call_all()
    
    security.declarePrivate('errback')
    def errback(self, fail):
        self.aq_inner.aq_parent.resolve(self)
        if self.raw_state != UNCALLED:
            # we don't want to abort the transaction because there's a good 
            # chance that we *need* to have the manager resolve this deferred.
            # Therefore, we return a failure rather than raise a transaction.
            # We do not call the errbacks.
            logging.getLogger('zasync').error(
                "Zope deferred %r has already been called" % (self.key,))
            return failure.Failure(
                RuntimeError("Deferred has already been called"))
        self.raw_state = FAILURE
        self.original_failure = self.failure = fail
        return self.__call_all()
    
    security.declareProtected(permissions.ModifyDeferred, 
                              'addCallback')
    def addCallback(self, callback):
        return self.addCallbacks(callback)
    
    security.declareProtected(permissions.ModifyDeferred, 
                              'addErrback')
    def addErrback(self, errback):
        return self.addCallbacks(errback=errback)
        
    def _convertArg(self, arg, default):
        if not arg or arg==default:
            arg = None
        else:
            if isinstance(arg, basestring):
                arg = Expression(arg)
            elif (isinstance(arg, (list, tuple)) and 
                  len(arg)==2 and 
                  isinstance(arg[0], basestring) and
                  isinstance(arg[1], basestring)):
                arg = (arg[0], Expression(arg[1]))
            else:
                raise ValueError(
                    "Must be None, expression string, or pair of (result name, "
                    "expression string)", arg)
        return arg
    
    security.declareProtected(permissions.ModifyDeferred, 
                              'addCallbacks')
    def addCallbacks(self, callback=None, errback=None):
        if errback is None and callback is None:
            raise ValueError("No callback or errback added")
        callback = self._convertArg(callback, 'result')
        errback = self._convertArg(errback, 'failure')
        self.__callbacks += ((callback, errback),)
        # call immediately if called
        if self.raw_state != UNCALLED:
            return self.__call_all()
        return self
    
    # XXX this part of the API may be in flux; Twisted has deprecated it
    security.declareProtected(permissions.ModifyDeferred, 
                              'setTimeout')
    def setTimeout(self, seconds):
        if self.getState() is not None:
            raise RuntimeError(
                "Cannot set timeout on completed deferred")
        seconds = int(seconds)
        if seconds < 0:
            raise ValueError("Seconds must be >= 0")
        if self.timeout < seconds:
            raise ValueError(
                "Cannot set timeout greater than previous value", self.timeout)
        self.timeout = seconds
    
    security.declareProtected(permissions.View, 'remainingSeconds')
    def remainingSeconds(self):
        """returns number of seconds until timeout (positive integer), or 
        number of seconds since timeout (negative integer)"""
        timeout = self.timeout
        end = self.creation_date + datetime.timedelta(seconds = timeout)
        now = datetime.datetime.now()
        delta = end - now
        if delta < datetime.timedelta():
            timeout = -(-delta).seconds # XXX should include day seconds too
        else:
            timeout = delta.seconds
        return timeout
InitializeClass(Deferred)

def getDeferredInfo(dictionary, sort_field=None, reverse=False):
    res = []
    for d in dictionary.values():
        plugin, args, kwargs = d.getSignature()
        state = d.getState()
        value = original_value = original_state = None
        if state is not None:
            value = d.getValue()
            if d.original_failure is None:
                original_value = d.original_result
                original_state = state_name_map[CALLED]
            else:
                original_value = d.original_failure
                original_state = state_name_map[FAILURE]
        info ={
            'key': d.key, 
            'user': d.getOwnerTuple(), 
            'plugin': plugin,
            'args': args,
            'kwargs': kwargs,
            'state': state,
            'value': value,
            'creation_date': d.creation_date,
            'resolution_date': d.resolution_date,
            'original_state': original_state,
            'original_value': original_value}
        if sort_field is None:
            res.append(info)
        else:
            res.append((info.get(sort_field), info))
    if sort_field is not None:
        res.sort()
        if reverse:
            res.reverse()
        res = [val for sort, val in res]
    return res

class AsynchronousCallManager(PropertyManager, SimpleItem):
    """a tool that holds deferreds for zasync to manipulate"""
    security = ClassSecurityInfo()
    
    manage_options = (
        ({'label':'Overview', 'action':'manage_overview',},
         {'label':'Calls', 'action':'manage_calls',},) +
        PropertyManager.manage_options
        + SimpleItem.manage_options)
    
    _properties = (
        {'id':'rotation_period', 'type': 'int', 'mode':'w',
         'label': 'Resolved cache rotation period in seconds'},
        {'id':'poll_interval', 'type': 'int', 'mode': 'w',
         'label': 'Interval between zasync call polls'},
        )
    
    id = 'asynchronous_call_manager'
    title = meta_type = 'Asynchronous Call Manager'
    icon = "misc_/zasync/tool.gif"
    rotation_period = 60*60*24 # a day
    poll_interval = 5
    __plugins = ()
    _next_rotate = _last_ping = _last_pong = None
    
    # keep webdav interface off deferreds and tool
    __implements__ = (interfaces.IZasyncAsynchronousCallManager,)
    
    def _getBrowserId(self):
        bim = getattr(self, BROWSERID_MANAGER_NAME)
        return bim.getBrowserId()
    
    def __init__(self, id=None):
        if id is not None:
            self.id = id
        # items to be picked up by zasync
        self._new = OOBTree.OOBTree()
        # items collected by zasync from the queue
        self._accepted = OOBTree.OOBTree()
        # long term cache
        self._resolved = bforests.OOBForest()
        self._next_rotate = None
        self._last_ping = None
        self._last_pong = None
    

    security.declareProtected(permissions.ViewManagementScreens, 
                              'manage_overview')
    manage_overview = PageTemplateFile(
        'www/controlAsynchronousCallManagerForm.zpt', globals(),
        __name__='manage_overview')
    

    security.declareProtected(permissions.ViewManagementScreens, 
                              'manage_calls')
    manage_calls = PageTemplateFile(
        'www/analyzeCalls.zpt', globals(),
        __name__='manage_calls')
    
    security.declareProtected(permissions.ViewManagementScreens, 
                              'ping')
    def ping(self, REQUEST=None):
        """make a ping request to the zasync client to see if it replies
        in the heartbeat method."""
        self._last_ping = datetime.datetime.now()
        self._last_pong = None
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect(
                '%s/manage_overview' % self.absolute_url())

    security.declareProtected(permissions.ViewManagementScreens, 
                              'getLastPing')
    def getLastPing(self):
        return self._last_ping

    security.declareProtected(permissions.ViewManagementScreens, 
                              'getLastPong')
    def getLastPong(self):
        return self._last_pong
    
    # property code copied from CMFCore/utils.py SimpleItemWithProperties >>>
    security.declarePrivate('manage_addProperty')
    security.declarePrivate('manage_delProperties')
    security.declarePrivate('manage_changePropertyTypes')

    def manage_propertiesForm(self, REQUEST, *args, **kw):
        'An override that makes the schema fixed.'
        my_kw = kw.copy()
        my_kw['property_extensible_schema__'] = 0
        return apply(PropertyManager.manage_propertiesForm,
                     (self, self, REQUEST,) + args, my_kw)

    security.declarePublic('propertyLabel')
    def propertyLabel(self, id):
        """Return a label for the given property id
        """
        for p in self._properties:
            if p['id'] == id:
                return p.get('label', id)
        return id
    # <<< end copy from CMF
    
    security.declarePrivate('resolve')
    def resolve(self, deferred):
        try:
            del self._accepted[deferred.key]
        except KeyError:
            pass # presumably resolved before
        else:
            self._resolved[deferred.key] = deferred
            if self._next_rotate is None and self.rotation_period:
                self._next_rotate = (
                    datetime.datetime.now() + 
                    datetime.timedelta(seconds=self.rotation_period))
    
    security.declarePublic('getNextCacheRotation')
    def getNextCacheRotation(self):
        return self._next_rotate
    
    security.declareProtected(permissions.ViewManagementScreens, 
                              'resetNextCacheRotation')
    def resetNextCacheRotation(self, clear=False):
        if clear:
            res = self._next_rotate = None
        else:
            res = self._next_rotate = (
                datetime.datetime.now() +
                datetime.timedelta(seconds=self.rotation_period))
        return res

    security.declarePrivate('setPlugins')
    def setPlugins(self, plugins): # protect plugins by groups or perms?
        "set item list, name: description"
        # XXX it would be good to include a signature validation function
        # XXX it would be good to be able to specify if some plugins are
        # not available to session calls--although that needs to be more
        # general; session call approach may need to be adjusted.  See XXX in
        # getSessionDeferred.
        self.__plugins = tuple(plugins)

    security.declareProtected(permissions.View, 'listPlugins')
    def listPlugins(self):
        """show plugin names available.  key is name, value is description.
        If no plugins, no engine should be expected (i.e., the tool will
        not work)."""
        return dict(self.__plugins)
    
    security.declarePrivate('_putCall')
    def _putCall(self, _id_prefix, _plugin, *args, **kwargs):
        if _plugin not in self.listPlugins(): # keys of dict
            raise ValueError(
                "Requested plugin is not registered")
        while 1:
            randomizer = key = random.randint(-sys.maxint-1, sys.maxint)
            if _id_prefix is not None:
                key = _id_prefix + (randomizer,)
            if self.getDeferred(key) is None:
                break
        args = sanitize(args)
        kwargs = sanitize(kwargs)
        d = Deferred(_plugin, args, kwargs)
        d.key = key
        d.id = repr(key)
        d.local_key = randomizer
        self._new[key] = d
        wrapped = d.__of__(self)
        wrapped.manage_fixupOwnershipAfterAdd()
        user=getSecurityManager().getUser()
        if user is not None:
            userid=user.getId()
            if userid is not None:
                wrapped.manage_setLocalRoles(userid, ['Owner'])
        return wrapped
    
    security.declareProtected(
        permissions.MakeAsynchronousApplicationCalls, 'putCall')
    def putCall(self, _plugin, *args, **kwargs):
        return self._putCall(
            None, _plugin, *args, **kwargs)
    
    security.declareProtected(
        permissions.MakeAsynchronousSessionCalls, 'putSessionCall')
    def putSessionCall(self, _plugin, *args, **kwargs):
        return self._putCall(
            (self._getBrowserId(),), _plugin, *args, **kwargs)
    
    security.declareProtected(
        permissions.MakeAsynchronousApplicationCalls, 'getDeferred')
    def getDeferred(self, d_id, default=None):
        for src in (self._new, self._accepted, self._resolved):
            res = src.get(d_id)
            if res is not None:
                return res.__of__(self)
        return default
    
    security.declareProtected(
        permissions.MakeAsynchronousSessionCalls, 'getSessionDeferred')
    def getSessionDeferred(self, d_id, default=None):
        # XXX this API doesn't support the use case of users who have limited
        # permissions but still should be able to store away keys for later
        # look-up, irrespective of sessions.  We should either have three 
        # levels of deferred call or let plugins specify permissions--but then
        # in what context?
        bid = self._getBrowserId()
        if isinstance(d_id, tuple):
            if d_id[0] != bid:
                return default
            d_id = d_id[-1]
        return self.getDeferred((bid, d_id), default)
    
    def __len__(self): # potentially expensive
        return len(self._new) + len(self._accepted) + len(self._resolved)

    def __nonzero__(self):
        return True
    
    security.declareProtected(
        permissions.ViewManagementScreens, 'listNewCalls')
    def listNewCalls(self, sort='creation_date', reverse=True):
        return getDeferredInfo(self._new, sort, reverse)
    
    security.declareProtected(
        permissions.ViewManagementScreens, 'listAcceptedCalls')
    def listAcceptedCalls(self, sort='creation_date', reverse=True):
        return getDeferredInfo(self._accepted, sort, reverse)
    
    security.declareProtected(
        permissions.ViewManagementScreens, 'listResolvedCalls')
    def listResolvedCalls(self, bucket=None, sort='creation_date', reverse=True):
        if bucket is None:
            d = self._resolved
        else:
            d = self._resolved.buckets[bucket]
        return getDeferredInfo(d, sort, reverse)
    
    security.declareProtected(
        permissions.ViewManagementScreens, 'lenNewCalls')
    def lenNewCalls(self):
        return len(self._new)
    
    security.declareProtected(
        permissions.ViewManagementScreens, 'lenAcceptedCalls')
    def lenAcceptedCalls(self):
        return len(self._accepted)
    
    security.declareProtected(
        permissions.ViewManagementScreens, 'lenResolvedCalls')
    def lenResolvedCalls(self, bucket=None):
        if bucket is None:
            d = self._resolved
        else:
            d = self._resolved.buckets[bucket]
        return len(d)
    
    security.declareProtected(
        permissions.ViewManagementScreens, 'lenResolvedBuckets')
    def lenResolvedBuckets(self):
        return len(self._resolved.buckets)
    
    security.declareProtected(
        permissions.ViewManagementScreens, 'nextResolvedBucketRotation')
    def nextResolvedBucketRotation(self):
        return self._next_rotate
    
    security.declarePrivate('acceptAll')
    def acceptAll(self):
        self._accepted.update(self._new)
        res = self._new.values()
        self._new.clear()
        return res
    
    security.declarePrivate('getAcceptedCalls')
    def getAcceptedCalls(self):
        return self._accepted.values()
    
    security.declarePrivate('heartbeat')
    def heartbeat(self):
        next_rotate = self._next_rotate
        now = datetime.datetime.now()
        if next_rotate is not None and next_rotate < now:
            self._resolved.rotateBucket()
            if self._resolved: # i.e., not empty; len is expensive!
                self._next_rotate = (
                    now + 
                    datetime.timedelta(seconds=self.rotation_period))
            else:
                self._next_rotate = None
        if self._last_ping is not None and self._last_pong is None:
            self._last_pong = now
InitializeClass(AsynchronousCallManager)

constructAsynchronousCallManagerForm = PageTemplateFile(
    'www/constructAsynchronousCallManagerForm.zpt', globals(),
    __name__='manage_addSchedulerForm')

def constructAsynchronousCallManager(
    dispatcher, id="asynchronous_call_manager", 
    poll_interval=None, rotation_period=None, RESPONSE=None):
    """Construct an AsynchronousCallManager"""
    acm = AsynchronousCallManager(id)
    if poll_interval is not None:
        acm.poll_interval = max(poll_interval, 2)
    if rotation_period is not None:
        acm.rotation_period = abs(rotation_period)
    container = dispatcher.this()
    container._setObject(id, acm)

    if RESPONSE is not None:
        RESPONSE.redirect(container.absolute_url() + '/manage_main')
