# Modifications of the standard library Queue tests to test the           
# bucketqueue; because I'm leveraging the Queue tests already in the      
# library, I'm using their (admittedly antiquated) approach.  Because I'm 
# importing test_queue from the original tests, and it runs automatically 
# on import, running the tests generates the following output (ideally):  
#
#   Simple Queue tests seemed to work
#   Failing Queue tests seemed to work
#   Simple BucketQueue tests seemed to work
#   Failing BucketQueue tests seemed to work
#
# The first two are simply running tests on the original Python Queue.  The 
# last two indicate success for the BucketQueue.

import threading, time
from test.test_queue import FailingQueueException
from test.test_queue import FailingQueueTest as OriginalFailingQueueTest
from test.test_queue import SimpleQueueTest as OriginalSimpleQueueTest
from test.test_support import verify, TestFailed, verbose
from Products.zasync import bucketqueue

queue_size = 5

# Execute a function that blocks, and in a seperate thread, a function that
# triggers the release.  Returns the result of the blocking function.
class _TriggerThread(threading.Thread):
    def __init__(self, fn, args, prefn=None, preargs=()):
        self.fn = fn
        self.args = args
        self.prefn = prefn
        self.preargs = preargs
        self.preEvent = threading.Event()
        self.startedEvent = threading.Event()
        threading.Thread.__init__(self)
    def run(self):
        global thread_res
        if self.prefn is not None:
            self.prefn(*self.preargs)
        self.preEvent.set()
        time.sleep(.1)
        self.startedEvent.set()
        self.result = self.fn(*self.args)

def _doBlockingTest(block_func, block_args, 
                    trigger_func, trigger_args, 
                    pre_func=None, pre_args=()):
    t = _TriggerThread(trigger_func, trigger_args, pre_func, pre_args)
    t.start()
    t.preEvent.wait()
    try:
        return block_func(*block_args)
    finally:
        # If we unblocked before our thread made the call, we failed!
        if not t.startedEvent.isSet():
            raise TestFailed("blocking function '%r' appeared not to block" %
                             block_func)
        t.join(1) # make sure the thread terminates
        if t.isAlive():
            raise TestFailed("trigger function '%r' appeared to not return" %
                             trigger_func)

# A BucketQueue subclass that can provoke failure at a moment's notice

class FailingQueue(bucketqueue.BucketQueue):
    def __init__(self, *args):
        self.fail_next_put = False
        self.fail_next_get = False
        bucketqueue.BucketQueue.__init__(self, *args)
    def _put(self, item):
        if self.fail_next_put:
            self.fail_next_put = False
            raise FailingQueueException, "You Lose"
        return bucketqueue.BucketQueue._put(self, item)
    def _removeIx(self, ix):
        if self.fail_next_get:
            self.fail_next_get = False
            raise FailingQueueException, "You Lose"
        return bucketqueue.BucketQueue._removeIx(self, ix)

def SeparateBucketFailingQueueTest(q):
    # this test has every item in a separate bucket
    for i in range(queue_size):
        q.makeBucket(i, silent=True)
    if not q.empty():
        raise RuntimeError, "Call this function with an empty queue"
    for i in range(queue_size-1):
        q.put(i, bucket=i)
    next_bucket = i + 1 # queue_size - 1
    # Test a failing non-blocking put.
    q.fail_next_put = True
    try:
        q.put("oops", block=0, bucket=next_bucket)
        raise TestFailed("The queue didn't fail when it should have")
    except FailingQueueException:
        pass
    q.fail_next_put = True
    try:
        q.put("oops", timeout=0.1, bucket=next_bucket)
        raise TestFailed("The queue didn't fail when it should have")
    except FailingQueueException:
        pass
    q.put("last", bucket=next_bucket)
    verify(q.full(), "Queue should be full")
    # Test a failing blocking put
    q.fail_next_put = True
    try:
        _doBlockingTest( q.put, ("full", True, None, next_bucket), q.get, ())
        raise TestFailed("The queue didn't fail when it should have")
    except FailingQueueException:
        pass
    # Check the Queue isn't damaged.
    # put failed, but get succeeded - re-add
    q.put("last", bucket=next_bucket)
    # Test a failing timeout put
    q.fail_next_put = True
    try:
        _doBlockingTest(q.put, ("full", True, 0.2, next_bucket), 
                        q.get, ())
        raise TestFailed("The queue didn't fail when it should have")
    except FailingQueueException:
        pass
    # Check the Queue isn't damaged.
    # put failed, but get succeeded - re-add
    q.put("last", bucket=next_bucket)
    verify(q.full(), "Queue should be full")
    q.get()
    verify(not q.full(), "Queue should not be full")
    q.put("last", bucket=next_bucket)
    verify(q.full(), "Queue should be full")
    # Test a blocking put
    _doBlockingTest( q.put, ("full", True, None, next_bucket), q.get, ())
    # Empty it
    for i in range(queue_size):
        q.get()
    verify(q.empty(), "Queue should be empty")
    q.put(0, "first")
    q.fail_next_get = True
    try:
        q.get()
        raise TestFailed("The queue didn't fail when it should have")
    except FailingQueueException:
        pass
    verify(not q.empty(), "Queue should not be empty")
    q.fail_next_get = True
    try:
        q.get(timeout=0.1)
        raise TestFailed("The queue didn't fail when it should have")
    except FailingQueueException:
        pass
    verify(not q.empty(), "Queue should not be empty")
    q.get()
    verify(q.empty(), "Queue should be empty")
    q.fail_next_get = True
    try:
        _doBlockingTest( q.get, (), q.put, ('empty', True, None, 0))
        raise TestFailed("The queue didn't fail when it should have")
    except FailingQueueException:
        pass
    # put succeeded, but get failed.
    verify(not q.empty(), "Queue should not be empty")
    q.get()
    verify(q.empty(), "Queue should be empty")

def SeparateBucketSimpleQueueTest(q):
    # this test has every item in a separate bucket
    for i in range(queue_size):
        q.makeBucket(i, silent=True)
    if not q.empty():
        raise RuntimeError, "Call this function with an empty queue"
    # I guess we better check things actually queue correctly a little :)
    q.put(111, bucket=1)
    q.put(222, bucket=2)
    verify(q.get() == 111 and q.get() == 222,
           "Didn't seem to queue the correct data!")
    for i in range(queue_size-1):
        q.put(i, i)
    next_bucket = i + 1
    verify(not q.full(), "Queue should not be full")
    q.put("last", bucket=next_bucket)
    verify(q.full(), "Queue should be full")
    try:
        q.put("full", block=0, bucket=next_bucket)
        raise TestFailed("Didn't appear to block with a full queue")
    except bucketqueue.Full:
        pass
    try:
        q.put("full", timeout=0.1, bucket=next_bucket)
        raise TestFailed("Didn't appear to time-out with a full queue")
    except bucketqueue.Full:
        pass
    # Test a blocking put
    _doBlockingTest( q.put, ("full", True, None, next_bucket), q.get, ())
    _doBlockingTest( q.put, ("full", True, 0.2, next_bucket), q.get, ())
    # Empty it
    for i in range(queue_size):
        q.get()
    verify(q.empty(), "Queue should be empty")
    try:
        q.get(block=0)
        raise TestFailed("Didn't appear to block with an empty queue")
    except bucketqueue.Empty:
        pass
    try:
        q.get(timeout=0.1)
        raise TestFailed("Didn't appear to time-out with an empty queue")
    except bucketqueue.Empty:
        pass
    # Test a blocking get
    _doBlockingTest(q.get, (), q.put, ('empty', True, None, 0))
    _doBlockingTest(q.get, (True, 0.2), q.put, ('empty', True, None, 0))

def BucketLockSimpleQueueTest(q):
    q.makeBucket(1, silent=True)
    q.makeBucket(2, silent=True)
    q.makeBucket(3, silent=True)
    if not q.empty():
        raise RuntimeError, "Call this function with an empty queue"
    q.put(111, bucket=1)
    verify(q.primed(), "Primed should be True")
    verify(q.available(), "Available should be True")
    q.put(112, bucket=1)
    q.put(113, bucket=1)
    verify(q.get()==111 and q.get()==112 and q.get()==113, 
           "Didn't get the correct queue data")
           # since it is all in the same thread, no blocks
    # now this thread has the bucket lock to 1.
    q.releaseBucket()
    q.put(111, bucket=1)
    q.put(112, bucket=1)
    q.put(113, bucket=1)
    verify(
        _doBlockingTest(
            q.get, (),
            q.put, (221, True, None, 2),
            q.get, ()) == 221, # pre gets 111
        "Didn't get the correct queue data")
    # now this thread holds the bucket lock to 2, and the dead thread holds
    # the bucket lock to 1
    verify(not q.primed(), "Primed should be False")
    verify(not q.empty(), "Empty should be False")
    verify(not q.available(), "Available should be False")
    try:
        q.get(block=0) # this releases the bucket lock on 2, but there's none
        # there
        raise TestFailed("Didn't appear to block with an empty queue")
    except bucketqueue.Empty:
        pass # XXX Queue.Empty isn't quite right anymore :-(
    verify(not q.primed(), "Primed should be False")
    verify(not q.available(), "Available should be False")
    q.put(222, bucket=2)
    verify(q.primed(), "Primed should be True")
    verify(q.available(), "Available should be True")
    verify(q.get()==222, "Didn't get the correct queue data")
    q.put(223, bucket=2)
    verify(not q.primed(), "Primed should be False") # (again) but since this
    # thread holds the lock on bucket 2, we can still get one.
    verify(q.available(), "Available should be True")
    verify(q.get()==223, "Didn't get the correct queue data")
    verify(q.releaseBucket(1)==1,
           "Release of bucket didn't work correctly")
    verify(q.releaseBucket()==2,
           "Release of bucket didn't work correctly")
    verify(
        _doBlockingTest(
            q.get, (),
            q.releaseBucket, (),
            q.get, ()) == 113, # pre gets 112
        "Didn't get the correct queue data")
    verify(q.empty(), "Queue should be empty")
    # thread holds the lock on 1
    verify(q.releaseBucket()==1,
           "Release of bucket didn't work correctly")
    verify(q.empty(), "Queue should be empty")

def test():
    q=bucketqueue.BucketQueue(queue_size)
    # Do it a couple of times on the same queue
    OriginalSimpleQueueTest(q)
    OriginalSimpleQueueTest(q)
    SeparateBucketSimpleQueueTest(q)
    SeparateBucketSimpleQueueTest(q)
    q = bucketqueue.BucketQueue()
    BucketLockSimpleQueueTest(q)
    BucketLockSimpleQueueTest(q)
    if verbose:
        print "Simple BucketQueue tests seemed to work"
    q = FailingQueue(queue_size)
    OriginalFailingQueueTest(q)
    OriginalFailingQueueTest(q)
    SeparateBucketFailingQueueTest(q)
    SeparateBucketFailingQueueTest(q)
    if verbose:
        print "Failing BucketQueue tests seemed to work"

if __name__ == "__main__":
    test()
