##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""zope test utilities

$Id: zopetestutils.py,v 1.1.1.1 2004/10/10 23:37:13 poster Exp $
"""
import os, doctest, new, unittest, StringIO
import ZODB
from ZODB.DB import DB
from ZEO.ClientStorage import ClientStorage
from ZODB.DemoStorage import DemoStorage
from OFS.Application import Application
from OFS.Application import initialize
from Testing.makerequest import makerequest
from AccessControl.User import system as ADMIN, nobody as ANONYMOUS
from AccessControl import getSecurityManager
from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManagement import setSecurityManager
from Acquisition import aq_parent, aq_inner

manager = None
connection = None

class CheckpointManager:

    # XXX TODO: remember past pops so we can skip over unnecessary undos.

    def __init__(self, connection=None, domain='localhost', port=8100):
        # a checkpoint manager can be instantiated either with a connection
        # (a database object's _p_jar) or a ZEO domain and port
        if connection is None:
            self.db = DB(ClientStorage((domain, port)))
            self.connection = self.db.open()
        else:
            self.connection = connection
            self.db = self.connection.db()
        self.root = self.connection.root()
        self.app = self.root['Application']
        self.undo_stack = []
    
    def push(self):
        get_transaction().commit()
        current_td = self.db.undoInfo(0,-1)[0]
        if self.undo_stack and self.undo_stack[-1]==current_td:
            raise RuntimeError("No change since list begin")
        self.undo_stack.append(current_td)
    
    def pop(self):
        get_transaction().commit() # XXX could be cleaned up
        undosteps = []
        checkpoint = self.undo_stack.pop()
        start = 0
        step = -20
        info = self.db.undoInfo
        while checkpoint is not None:
            history = info(start, start+step)
            if not history:
                raise RuntimeError("Checkpoint not found.")
            for i in history:
                if i['id']==checkpoint['id']:
                    checkpoint = None
                    break
                if i['time'] < checkpoint['time']:
                    raise RuntimeError(
                        "Checkpoint not found.")
                undosteps.append(i['id'])
            start += step
        undo = self.db.undo
        for tid in undosteps:
            undo(tid)
        get_transaction().commit()
        self.connection.sync() # XXX needed?
    
    def close(self):
        self.connection.close()

def forget():
    global manager
    manager.pop()

def remember():
    global manager
    manager.push()

def retry():
    global manager
    manager.pop()
    manager.push()

def app():
    global connection, manager
    if connection is None:
        db = ZODB.DB(DemoStorage(quota=(1<<20)))
        connection = db.open()
        application = Application()
        connection.root()["Application"] = application
        get_transaction().commit(1)
        responseOut = StringIO.StringIO()
        app = makerequest(application, stdout=responseOut)        
        initialize(app)
        manager = CheckpointManager(connection)
        remember()
    return connection.root()["Application"]

old_securitymanager = None

def startRequest(user=ANONYMOUS, previous=None, out=None):
    """pass a user object or a user tuple of (path to user database, user id),
    or defaults to the anonymous user; pass a previous request object to
    carry over cookie information; pass an open file to which the RESPONSE
    should write or a new StringIO will be created (accessible as 
    REQUEST.RESPONSE.stdout)"""
    global old_securitymanager
    if out is None:
        out = StringIO.StringIO()
    application = makerequest(app(), stdout=out)
    if isinstance(user, (tuple, list)):
        user_db, user_id = user
        user_db = application.unrestrictedTraverse(user_db, None)
        if user_db is None:
            user = ANONYMOUS
        else:
            user = user_db.getUserById(user_id, ANONYMOUS)
        user = user.__of__(user_db)
    if aq_parent(user) is None:
        user = user.__of__(application)
    if old_securitymanager is None:
        old_securitymanager = getSecurityManager()
    newSecurityManager(application.REQUEST, user)
    if previous is not None:
        # try to keep the cookies around
        cookies = previous.cookies
        for name, info in previous.RESPONSE.cookies.items():
            if info.get('expires') == "Sun, 10-May-1971 11:59:00 GMT":
                # XXX should really be more careful, checking to see if the
                # date is prior to now
                cookies.pop(name, None)
            else:
                cookies[name] = info.get('value')
        application.REQUEST.cookies.update(cookies)
    return application

def closeRequest():
    global old_securitymanager, connection
    get_transaction().commit()
    setSecurityManager(old_securitymanager)
    old_securitymanager = None

def anotherRequest(app, out=None):
    user = getSecurityManager().getUser()
    base_user = aq_parent(user)
    if base_user is ANONYMOUS or base_user is ADMIN:
        user = base_user
    else:
        user = (aq_parent(aq_inner(user)).getPhysicalPath(), user.getId())
    previous = app.REQUEST
    closeRequest()
    return startRequest(user, previous, out)

# DocFileSuite is slightly modified from original in Zope 3 from March 2004:

def DocFileSuite(*paths, **globals):
    globals['__name__'] = '__main__'
    t = doctest.Tester(globs=globals)
    suite = unittest.TestSuite()
    dir = os.path.split(__file__)[0]
    for path in paths:
        path = os.path.join(dir, path)
        source = open(path).read()
        def runit(path=path, source=source):
            doctest._utest(t, path, source, path, 1)
        runit = new.function(runit.func_code, runit.func_globals, path,
                             runit.func_defaults, runit.func_closure)
        f = unittest.FunctionTestCase(runit,
                                      description="doctest from %s" % path)
        suite.addTest(f)
    return suite

def StringGenerator(src='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'):
    "infinite-ish unique string generator"
    for el in src:
        yield el
    for pre in StringGenerator(src):
        for el in src:
            yield pre + el

def NumberGenerator(number=0, interval=1):
    "infinite-ish unique number generator"
    while 1:
        yield number
        number += interval

def test_stuff():
    application = app()
    global manager
    try:
        manager.push()
        application.manage_addFolder('foo')
        get_transaction().commit()
        application.foo.manage_addFile('file')
        get_transaction().commit()
        assert application.foo.file.getId()=='file'
        manager.pop()
        try:
            application.foo
        except AttributeError:
            return "yay"
        else:
            return "boo"
    finally:
        manager.close()

if __name__=="__main__":
    print test_stuff()
