##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""zasync plugins

$Id: plugins.py,v 1.7 2004/12/06 20:20:09 poster Exp $
"""

import logging
from twisted.python import failure
from twisted.internet import reactor, defer
from zasync import client

#### simple schedule

def schedule(seconds):
    """proof of concept and "Hello World"; use to fire your callbacks after 
    seconds (approximately).  zope_exec is better for potentially 
    expensive tasks because it is cancellable (can time out) and you can guage
    better if the expensive task has started.  You might schedule a zope_exec 
    within a schedule callback, if you needed a scheduled expensive task."""
    d = defer.Deferred()
    reactor.callLater(seconds, d.callback, seconds)
    return d

#### aggegate plugins

def aggregatePlugins(zopeDeferredTuple, *calls):
    # XXX remove timeout feature; use the schedule plugin!
    # XXX also? instead? incorporate easy and introspectable chaining into
    # zopeDeferreds
    """aggregate calls to other plugins.  each call may either be a number
    (integer or float) of seconds to pause, or a tuple of (plugin name, args 
    tuple, kwargs dict).  For example, 
    
    aggregatePlugins(
        ('zope_exec', ('/my_site', 'home/my_script'), {}),
        5,
        ('zope_exec', ('/my_site', 'home/my_other_script'), {}))
    
    would ask zope_exec to call my_script in one transaction, wait 5 seconds,
    and then call my_other_script.
    
    If any of the plugins fail, the failure is returned without proceeding 
    further down the remaining calls.  The failure is annotated with a list of
    the completed calls ('completed_calls'), the call active during the failure
    ('active_call'), and remaining calls('remaining_calls').
    """
    thunkmaker = AggregateThunkMaker(zopeDeferredTuple, *calls)
    reactor.callLater(0, thunkmaker.makeCall)
    return thunkmaker.deferred

class AggregateThunkMaker(object):
    
    def __init__(self, zopeDeferredTuple, *calls):
        self.deferred = defer.Deferred()
        self.pending = list(calls)
        self.completed = []
        self.begun = None
        self.results = []
        self.path, self.zopeDeferredId = zopeDeferredTuple

    def makeCall(self):
        if not self.pending:
            self.deferred.callback(self.results)
        else:
            if self.begun is not None:
                self.completed.append(self.begun)
            task = self.pending.pop(0)
            self.begun = task
            if isinstance(task, (int, float)):
                reactor.callLater(task, self.makeCall)
            else:
                try:
                    plugin, args, kwargs = task
                except (ValueError, TypeError):
                    self.deferred.errback(failure.Failure())
                else:
                    try:
                        call_info = client.plugins[plugin]
                    except KeyError:
                        self.deferred.errback(failure.Failure())
                    else:
                        timeout = call_info["timeout"]
                        remainingSeconds = client.active[self.zopeDeferredId][0]
                        client.makeCall(
                            self.path, self.zopeDeferredId, plugin, args, kwargs, 
                            remainingSeconds, timeout, self.returnResult)
            
    def returnResult(self, value, path, zopeDeferredId, error=False):
        if error:
            value.completed_calls = tuple(self.completed)
            value.active_call = self.begun
            value.remaining_calls = tuple(self.pending)
            self.deferred.errback(value)
        else:
            self.results.append(value)
            reactor.callLater(0, self.makeCall)

#### LDAP, protected (with SSL) and unprotected

try:
    import ldap
except ImportError:
    pass
else:
    def query_unprotected_ldap(uri, base, scope, filterstr, attrlist=None): 
        """Query unprotected ldap (i.e., not ldaps).  arguments are the URI of
        the ldap server, the base of the query, the scope of the search, and 
        the query string.  The scope should be one of the following three
        strings:
        - 'object' (equivalent to python ldap's SCOPE_BASE constant)
        - 'subtree' (SCOPE_SUBTREE)
        - 'children' (SCOPE_ONELEVEL)
        
        example values:
        'ldap://ldap.example.edu', 'dc=example,dc=edu', 'subtree', query
        
        Failures are exceptions as defined by the python ldap package.
        """
        scope = {
            'object':ldap.SCOPE_BASE,
            'subtree':ldap.SCOPE_SUBTREE, 
            'children':ldap.SCOPE_ONELEVEL}.get(scope)
        if scope is None:
            raise ValueError(
                "scope must be one of 'object', 'subtree', or 'children'")
        l = ldap.initialize(uri)
        # protocol version 3 is preferred (do I really care?  not sure)
        try:
            l.set_option(ldap.OPT_PROTOCOL_VERSION, ldap.VERSION3)
        except ldap.LDAPError:
            l.set_option(ldap.OPT_PROTOCOL_VERSION, ldap.VERSION2)
        msgid = l.search(base, scope, filterstr, attrlist)
        d = defer.Deferred()
        d.addErrback(abandon_query, l, msgid)
        reactor.callLater(2, poll_results, d, l, msgid)
        return d
    
    def query_protected_ldap(uri, username, password, base, scope, filterstr,
                             attrlist=None): 
        """Query protected ldap (i.e., ldaps, over SSL).  arguments are the URI
        of the ldap server, the user name (including ldap qualifiers, as in the 
        example below), the user password, the base of the query, the scope of 
        the search, and the query string.  The scope should be one of the 
        following three strings:
        - 'object' (equivalent to python ldap's SCOPE_BASE constant)
        - 'subtree' (SCOPE_SUBTREE)
        - 'children' (SCOPE_ONELEVEL)
        
        example values ('ldapuser' and 'ldappass' are username and password 
        # values):
        'ldaps://ldap.example.edu:636', 
        'uid=ldapuser,ou=users,ou=special,dc=example,dc=edu', 'ldappass', 
        'dc=example,dc=edu', 'subtree', query
        
        Failures are exceptions as defined by the python ldap package.
        """
        scope = {
            'object':ldap.SCOPE_BASE,
            'subtree':ldap.SCOPE_SUBTREE, 
            'children':ldap.SCOPE_ONELEVEL}.get(scope)
        if scope is None:
            raise ValueError(
                "scope must be one of 'object', 'subtree', or 'children'")
        l = ldap.initialize(uri)
        # protocol version 3 is preferred (do I really care?  not sure)
        try:
            l.set_option(ldap.OPT_PROTOCOL_VERSION, ldap.VERSION3)
        except ldap.LDAPError:
            l.set_option(ldap.OPT_PROTOCOL_VERSION, ldap.VERSION2)
        l.simple_bind(username, password)
        msgid = l.search(base, scope, filterstr, attrlist)
        d = defer.Deferred()
        d.addErrback(abandon_query, l, msgid)
        reactor.callLater(2, poll_results, d, l, msgid)
        return d
    
    def abandon_query(failure, connection, msgid):
        logging.getLogger('zasync.plugins').info('abandoning ldap search')
        connection.abandon(msgid)
        return failure
    
    def poll_results(deferred, connection, msgid):
        if not deferred.called:
            logging.getLogger('zasync.plugins').debug('polling ldap')
            try:
                res = connection.result(msgid, timeout=0) # poll
            except:
                deferred.errback(failure.Failure())
            else:
                if res is None or res == (None, None):
                    reactor.callLater(2, poll_results, deferred, connection, msgid)
                else:
                    deferred.callback(res)

#### Zope Exec

import Queue, thread, time, sys, StringIO
from ZODB.POSException import ConflictError
from ZEO.Exceptions import ClientDisconnected
from Acquisition import aq_parent, aq_inner
from AccessControl.SecurityManagement import newSecurityManager
from Products.PageTemplates.Expressions import getEngine, SecureModuleImporter

from Products.zasync.manager import Expression, sanitize
from Products.zasync.bucketqueue import BucketQueue

MAXTHREADPOOL = 5

threadIds = []
threadPoolLock = thread.allocate_lock()
taskQueue = BucketQueue()
taskStatusLock = thread.allocate_lock()
taskStatus = {}
callbacks = Queue.Queue()
serverDown = False

CANCEL = object()
CALLBACK = 'CALLBACK'
ERRBACK = 'ERRBACK'

def getTaskStatus(key, del_if_cancel=False):
    taskStatusLock.acquire()
    try:
        val = taskStatus.get(key)
        if del_if_cancel and val is CANCEL:
            try:
                del taskStatus[key]
            except KeyError:
                pass
    finally:
        taskStatusLock.release()
    return val

def popTaskStatus(key):
    taskStatusLock.acquire()
    try:
        val = taskStatus.get(key)
        try:
            del taskStatus[key]
        except KeyError:
            pass
    finally:
        taskStatusLock.release()
    return val
        

def setTaskStatus(key, value):
    taskStatusLock.acquire()
    try:
        taskStatus[key] = value
    finally:
        taskStatusLock.release()
    

# zopeDeferredTuple is (tool path, zope deferred id)

def zope_exec(zopeDeferredTuple, homepath, action, *chain):
    """Perform a callback chain asynchronously, beginning with a single
    action, where an action is either a TALES expression or a tuple of (name,
    TALES expression).  If the action is a tuple, then the result of the
    expression is available to following expressions as results[name] (or
    results/$name).  Each link in the optional following callback chain may be
    a dictionary with one or both of the keys CALLBACK or ERRBACK and values
    that are actions, or may be an action, which will be interpreted as a
    CALLBACK.  The result of the final action is the return value to the
    deferred callback.  If a CALLBACK or ERRBACK is missing, it is a no-op: if
    a value for CALLBACK is needed but not provided in a dictionary, it is
    effectively 'result'; if a value for ERRBACK is needed but not provided in
    a dictionary, it is effectively 'failure'.
    
    homepath is a path from the root (which may be a boolean False value such 
    as None, a tuple or list of path elements, or a string, but will be 
    normalized to a string) to an object that will be available as "home" in the
    action expressions, and is also used as a domain so that only a single 
    zope_exec within a given homepath will be executed at a time.  Note that 
    nested domains are effectively completely different domains: a zope_exec
    with a "/" homepath may be attempted at the same time as a zope_exec 
    with a "/cmf_site" homepath.
    
    Actions are performed as (and in the security context of) the user who owns 
    the zope deferred.  A dummy request is available.
    
    The code checks between each action and immediately prior to committing to
    see if the deferred has been cancelled (timed out); if it has been 
    cancelled, the job is aborted.  zope_exec are not called (although errbacks
    on the zope deferred will be called, as usual).
    
    expressions have the following names available to them:
    
    - nothing: None
    - user: the user object (the owner of the zope deferred)
    - home: the object referred to by the homepath
    - userhome: the location of the user's acl_users folder
    - tool: the asynchronous manager
    - deferred: the Zope deferred
    - root: the physical Zope root
    - modules: the standard Zope secure module importer
    - result: the result of the most recent previous successful call 
              (initializes to None)
    - failure: the failure of the most recent previous unsuccessful call
               (initializes to None)
    - results: a dictionary of results returned from past actions, as described
               above.
    - request: the faux request
    
    Example of putting a call:
    
    root.asynchronous_call_manager.putCall(
        "zope_exec", # name of plugin
        "/another_site", # homepath
        'user/getId', # initial action
        'python: foo', # ...following callbacks and errbacks...
        {ERRBACK: 'python: failure.trap(modules["AccessControl"].Unauthorized, '
                  'root.special_script)'},
        {CALLBACK: ('portal', 'root/my_site'),
         ERRBACK: ('portal', 'nothing')}
        )

    """
    global taskQueue
    # normalize homepath
    if not homepath or homepath=="/":
        homepath = ''
    else:
        # this may raise errors: if so, makeCall will convert to failures
        homepath = normalize_path(homepath)
    taskQueue.makeBucket(homepath, silent=True)
    chain = (action,) + chain
    d = defer.Deferred()
    setTaskStatus(zopeDeferredTuple, d) # == pending
    logger = logging.getLogger('zasync.plugins')
    logger.debug(
        'zope_exec: scheduling zope deferred %r',
        zopeDeferredTuple)
    d.addErrback(abandon_zope_exec, zopeDeferredTuple)
    reactor.callLater(
        0, put_call, (zopeDeferredTuple, homepath, chain), homepath)
    return d

def put_call(call, bucket):
    logger = logging.getLogger('zasync.plugins')
    logger.debug(
        'zope_exec: putting call %r in queue for bucket (homepath) %r',
        call, bucket)
    taskQueue.put(call, bucket=bucket) # this must be after all of the work in
    # zope_exec, above, or else the data structures the worker thread expects
    # are not (necessarily) in place yet.  This is in a separate reactor call
    # just to free the reactor up a bit to do other work.
    reactor.callLater(0.05, start_worker)
    # this must be after the put, because start_worker only
    # creates a thread if it can see that one is necessary

def normalize_path(path):
    if not isinstance(path, (list, tuple)):
        path = path.split('/')
    tmp = []
    for el in path:
        if el=="..":
            if tmp:
                tmp.pop()
        elif el==".":
            pass
        else:
            tmp.append(el)
    if not tmp or tmp[0]:
        tmp.insert(0, '')
    return '/'.join(tmp)

def start_worker():
    global threadIds
    logger = logging.getLogger('zasync.plugins')
    threadPoolLock.acquire()
    try:
        len_threadIds = len(threadIds)
    finally:
        threadPoolLock.release()
    if len_threadIds < MAXTHREADPOOL and taskQueue.primed():
        new_id = thread.start_new_thread(zope_exec_worker, ())
        threadPoolLock.acquire()
        try:
            threadIds.append(new_id)
        finally:
            threadPoolLock.release()
        logger.debug(
            "zope_exec: added new worker thread  %r to thread pool "
            "(%d worker threads total)",
            new_id, len_threadIds + 1)
        for c in client.chores:
            if c[0] is schedule_mainthread:
                break
        else:
            client.chores.append((schedule_mainthread, (), {}))
            logger.debug(
                'zope_exec: added schedule_mainthread to zasync chore list')
        if len_threadIds + 1 < MAXTHREADPOOL and taskQueue.primed():
            reactor.callLater(0.1, start_worker) # if start_worker is called by 
            # scheduleServerRetry as set up in schedule_mainthread, then it is
            # possible that more threads could be needed.  We check here, and
            # if it seems we could use some more, then we schedule another 
            # thread to possibly start in a tenth of a second (giving active
            # threads a healthy chance to take a stab at the queue before we
            # look again)
    logger.debug('out of start_worker')

def abandon_zope_exec(failure, zopeDeferredTuple):
    """this is an errback for the twisted deferred that makes sure that a
    possibly-still-in-queue action chain is cancelled."""
    if getTaskStatus(zopeDeferredTuple) is not None:
        setTaskStatus(zopeDeferredTuple, CANCEL)
        logging.getLogger('zasync.plugins').debug(
            'zope_exec: attempting to cancel pending %r for worker thread',
            zopeDeferredTuple)
    return failure

def schedule_mainthread():
    # do jobs that the zope_exec_worker needs done in the main
    # thread.  Basically needed for letting zasync return failures or results
    # to Zope.  The housekeeping function in the client module kicks this 
    # regularly.
    global callbacks, serverDown
    logger = logging.getLogger('zasync.plugins')
    while 1:
        try:
            deferred, result = callbacks.get(block=False)
        except Queue.Empty:
            break
        else:
            if not deferred.called:
                if isinstance(result, failure.Failure):
                    logger.debug(
                        'zope_exec: scheduling an errback through zasync\n%s',
                        result.getErrorMessage())
                    deferred.errback(result)
                else:
                    logger.debug(
                        'zope_exec: scheduling a callback through zasync\n%s',
                        result)
                    deferred.callback(result)
    if serverDown:
        serverDown = False
        client.scheduleServerRetry(start_worker)
        logging.getLogger('zasync.plugins').debug(
            'zope_exec: scheduling start_worker to resume after the ZEO server '
            'returns')

def zope_exec_worker():
    """the thread method."""
    # XXX generalize this to not require zope deferred so other plugins can use 
    logger = logging.getLogger('zasync.plugins')
    thread_id = thread.get_ident()
    logger.debug(
        'zope_exec: beginning worker thread %r', thread_id)
    global taskQueue, taskStatus, threadIds, callbacks, serverDown
    from zasync.client import app, max_conflict_resolution_attempts
    application = None
    try:
        while 1: # keep on looking for tasks
            zopeDeferredTuple, homepath, actions = taskQueue.get() # blocks
            logger.debug(
                'zope_exec: thread %r got work from zope deferred %r',
                thread_id, zopeDeferredTuple)
            if getTaskStatus(zopeDeferredTuple, True) is CANCEL:
                logger.debug(
                    'zope_exec: zope deferred %r has been cancelled; '
                    'worker thread aborting', zopeDeferredTuple)
                continue
            toolpath, zopeDeferredId = zopeDeferredTuple
            for attempt in range(max_conflict_resolution_attempts):
                try:
                    try:
                        deferred = None
                        try:
                            sync = client.DB._storage.sync # important
                        except AttributeError:
                            pass
                        else:
                            sync()
                        application = client.app()
                        home = root = client.getRequestApp(application)
                        tool = root.unrestrictedTraverse(toolpath)
                        zopeDeferred = tool.getDeferred(zopeDeferredId)
                        user = zopeDeferred.getWrappedOwner()
                        newSecurityManager(None, user)
                        results = {}
                        if homepath: # may raise unauthorized or others;
                            # should be caught in bare except below
                            home = root.restrictedTraverse(homepath)
                        result = None
                        fail = None
                        success = True
                        context_dict = {
                            'nothing': None, # automatic; here for clarity
                            'user': user,
                            'home': home,
                            'here': home,
                            'userhome': aq_parent(
                                aq_inner(aq_parent(aq_inner(user)))),
                            'tool': aq_parent(aq_inner(zopeDeferred)),
                            'root': root,
                            'deferred': zopeDeferred,
                            'modules': SecureModuleImporter,
                            'request': root.REQUEST,
                            'result': result,
                            'failure': fail,
                            'results': results,}
                        for action in actions:
                            logger.debug(
                                'zope_exec: worker %r processing action %r',
                                thread_id, action)
                            call = name = None
                            if isinstance(action, basestring):
                                if success:
                                    call = action
                            elif isinstance(action, tuple):
                                if success:
                                    name, call = action
                            elif isinstance(action, dict):
                                if success:
                                    action = action.get(CALLBACK)
                                else:
                                    action = action.get(ERRBACK)
                                if isinstance(action, tuple):
                                    name, call = action
                                elif isinstance(action, basestring):
                                    call = action
                                elif action is not None:
                                    raise ValueError(
                                        "an action within a dictionary must be "
                                        "a string or a tuple pair", action)
                            else:
                                raise ValueError(
                                    "each action must be a string, a tuple "
                                    "pair, or a dict", action)
                            if (not call or 
                                success and call=='result' or
                                not success and call=='failure'): # no-ops
                                continue
                            if getTaskStatus(zopeDeferredTuple, True) is CANCEL:
                                logger.debug(
                                    'zope_exec: worker %r got a cancel request: '
                                    'aborting', thread_id)
                                raise defer.TimeoutError('Timed out.')
                            context = getEngine().getContext(context_dict)
                            call = Expression(call)
                            try:
                                res = call(context)
                            except ((ConflictError, KeyboardInterrupt, 
                                     SystemExit, ClientDisconnected)):
                                raise
                            except:
                                res = failure.Failure()
                            else:
                                if name is not None:
                                    results[name] = res
                            if isinstance(res, failure.Failure):
                                res.cleanFailure() # remove object references
                                fail = context_dict["failure"] = res
                                success = False
                            else:
                                result = context_dict["result"] = res
                                success = True
                        deferred = getTaskStatus(zopeDeferredTuple)
                        if deferred is not CANCEL:
                            if success:
                                result = sanitize(result)
                                get_transaction().commit()
                                logger.debug(
                                    'zope_exec: worker %r committed successful '
                                    'transaction',
                                    thread_id)
                            else:
                                get_transaction().abort()
                                msg = ('zope_exec: worker %r aborted failed '
                                       'transaction.' % (thread_id,))
                                if client.verbose_traceback:
                                    out = StringIO.StringIO()
                                    res.printDetailedTraceback(out)
                                    logger.debug(
                                        '%s\n\n%s\n\n%s',
                                        msg, out.getvalue(), 
                                        res.getErrorMessage())
                                else:
                                    logger.debug(msg, exc_info=True)
                        else:
                            get_transaction().abort()
                            logger.debug(
                                'zope_exec: worker %r got a cancel request at '
                                'the last possible moment: aborted.',
                                thread_id)
                            break # the timeout errback is supposed to have
                            # been communicated within the main zasync thread
                    except ConflictError:
                        get_transaction().abort() 
                        logger.debug(
                            'zope_exec: worker %s got conflict error', 
                            thread_id, exc_info=True)
                        if attempt < max_conflict_resolution_attempts:
                            time.sleep(attempt + 1) # XXX better idea?
                        # (and continue, retrying if appropriate)
                    except (KeyboardInterrupt, SystemExit):
                        get_transaction().abort() 
                        raise
                    except ClientDisconnected:
                        get_transaction().abort()
                        logger.debug(
                            "zope_exec: worker %s lost ZEO server!", thread_id)
                        if getTaskStatus(zopeDeferredTuple) is not CANCEL:
                            # retry later
                            logging.debug(
                                'zope_exec: worker %s rescheduling %r',
                                thread_id, zopeDeferredTuple)
                            taskQueue.put(
                                (zopeDeferredTuple, homepath, actions),
                                bucket=taskQueue.releaseBucket())
                            zopeDeferredTuple = None # foils the last finally.
                        # setting this module global communicates to the 
                        # schedule_mainthread chore that the start_worker
                        # function will need to be called once the server is 
                        # back online again
                        serverDown = True
                        raise
                    except:
                        # in other code, we often try to be nice even if the
                        # tool disappears for a moment.  We don't bother here,
                        # which lets the normal zasync code try to communicate 
                        # the failure back to the log (and Zope, if the tool
                        # and the associated deferred happen to magically 
                        # reappear).
                        get_transaction().abort()
                        f = failure.Failure()
                        out = None
                        if client.verbose_traceback:
                            out = StringIO.StringIO()
                            f.printDetailedTraceback(out)
                            logger.debug(
                                'zope_exec: worker %s got an exception.'
                                '\n\n%s\n\n%s',
                                thread_id, out.getvalue(), f.getErrorMessage())
                        else:
                            logger.debug(
                                'zope_exec: worker %s got an exception',
                                thread_id, exc_info=True)
                        if deferred is None:
                            deferred = popTaskStatus(zopeDeferredTuple)
                        if deferred is not CANCEL: # one last chance
                            if deferred is None:
                                if out is None:
                                    out = StringIO.StringIO()
                                    f.printDetailedTraceback(out)
                                    out = "\n\n%s\n\n" % (
                                        out.getvalue(), f.getErrorMessage())
                                else:
                                    out = ""
                                logger.error(
                                    'zope_exec: worker %s cannot find the task '
                                    'status for %r, so it cannot schedule '
                                    'zasync to pass a failure back to Zope.  '
                                    'This is a significant problem.  Here is a '
                                    'lot of diagnostic information:\n\n%s',
                                    thread_id, zopeDeferredTuple, out)
                            else:
                                callbacks.put((deferred, f))
                                logging.debug(
                                    'zope_exec: worker %s scheduled zasync to '
                                    'pass a failure back to Zope', thread_id)
                        break
                    else: # yay! it worked
                        # ok, so the the success or failure that we got from
                        # zope_exec here should be passed back to the deferred.
                        # we want to leverage the ConflictError handling in the
                        # main zasync process, so we need to pass the result
                        # back to the main thread for zasync to communicate.
                        # all work in Zope should already be committed, so we
                        # shouldn't have to mess further with the transaction.
                        deferred = popTaskStatus(zopeDeferredTuple)
                        if deferred is None:
                            f = failure.Failure()
                            out = StringIO.StringIO()
                            f.printDetailedTraceback(out)
                            out = out.getvalue()
                            logger.error(
                                'zope_exec: worker %s cannot find the task '
                                'status for %r, so it cannot schedule '
                                'zasync to pass a result back to Zope.  '
                                'This is a significant problem.  Here is a lot '
                                'of diagnostic information:\n\n%s',
                                thread_id, zopeDeferredTuple, out)
                        elif success:
                            logger.debug(
                                'zope_exec: worker %s get final result %r; '
                                'scheduling zope deferred callback', 
                                thread_id, result)
                            callbacks.put((deferred, result))
                        else:
                            logger.debug(
                                'zope_exec: worker %s get final failure (%s); '
                                'scheduling zope deferred errback', 
                                thread_id, fail.getErrorMessage())
                            callbacks.put((deferred, fail))
                        break
                finally:
                    if application is not None:
                        application._p_jar.close()
                        application = None
            else: # too many conflict resolution attempts
                logger.debug(
                    'zope_exec: worker %s got too many conflict errors: '
                    'giving up.', thread_id)
                deferred = popTaskStatus(zopeDeferredTuple)
                if deferred is not None and deferred is not CANCEL:
                    logger.debug(
                        'zope_exec: worker %s scheduling failure message with '
                        'zasync', thread_id)
                    callbacks.put(
                        (deferred, failure.Failure()))
    finally:
        logger.debug(
            'zope_exec: worker %s cleaning up and going away', thread_id)
        taskQueue.releaseBucket() # make sure I don't lock a bucket forever
        threadPoolLock.acquire()
        try:
            threadIds.remove(thread.get_ident()) # tell start_worker I'm dead
        finally:
            threadPoolLock.release()
        deferred = getTaskStatus(zopeDeferredTuple, True)
        if deferred is not None: # looks like we still need to clean up the 
                                 # deferred we were just working on.
            if deferred is not CANCEL:
                logger.debug(
                    'zope_exec: worker %s scheduling failure message with '
                    'zasync', thread_id)
                callbacks.put(
                    (deferred, failure.Failure()))
